__author__ = "Simon Nilsson"

import glob
import itertools
import os
import webbrowser
from tkinter import *
from typing import List, Union

from simba.cue_light_tools.cue_light_analyzer import CueLightAnalyzer
from simba.cue_light_tools.cue_light_clf_statistics import CueLightClfAnalyzer
from simba.cue_light_tools.cue_light_movement_statistics import \
    CueLightMovementAnalyzer
from simba.cue_light_tools.cue_light_visualizer import CueLightVisualizer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.pop_ups.cue_light_clf_analyzer_popup import \
    CueLightClfAnalyzerPopUp
from simba.ui.pop_ups.cue_light_data_analyzer_popup import \
    CueLightDataAnalyzerPopUp
from simba.ui.pop_ups.cue_light_movement_analyzer_popup import \
    CueLightMovementAnalyzerPopUp
from simba.ui.pop_ups.cue_light_visualizer_popup import CueLightVisulizerPopUp
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, SimbaButton,
                                        SimBADropDown, SimBALabel)
from simba.utils.checks import check_float, check_int
from simba.utils.enums import Keys, Links
from simba.utils.errors import (CountError, NoChoosenClassifierError,
                                NoFilesFoundError, NoROIDataError)
from simba.utils.read_write import (find_video_of_file, get_all_clf_names,
                                    get_fn_ext, read_config_entry)


class CueLightMainPopUp(ConfigReader, PopUpMixin):
    """
    Launch cue light analysis GUI in SimBA.

    :parameter str config_path: path to SimBA project config file in Configparser format

    .. note::
       `Cue light tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/cue_light_tutorial.md>`__.


    :examples:
    >>> test = CueLightMainPopUp(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini")
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path)
        if len(self.outlier_corrected_paths) == 0:
            raise NoFilesFoundError(msg=f'No data found in the {self.outlier_corrected_dir} directory. Cannot compute cue-lights statistics.', source=self.__class__.__name__)
        if not os.path.isfile(self.roi_coordinates_path):
            raise NoROIDataError(msg=f'No ROIs found in project ({self.roi_coordinates_path} file not found). Draw ROIs before computing cue-light statistics', source=self.__class__.__name__)
        self.read_roi_data()
        PopUpMixin.__init__(self, size=(750, 300), title="SIMBA CUE LIGHT ANALYZER", icon='light_bulb')
        self.max_len, self.cue_light_dropdowns = max(len(s) for s in self.roi_names), []
        self._get_cue_lights_frm()

    def _get_cue_lights_frm(self):
        def _get_cue_light_names(cue_light_cnt: int = 1):
            for i in self.cue_light_dropdowns:
                i.destroy()
            self.cue_light_dropdowns = []
            for j in range(cue_light_cnt):
                cue_light_dropdown = SimBADropDown(parent=self.cue_light_settings_frm, dropdown_options=self.roi_names, label=f'CUE LIGHT {j+1}', label_width=20, dropdown_width=self.max_len + 2, value=self.roi_names[j])
                self.cue_light_dropdowns.append(cue_light_dropdown)
                cue_light_dropdown.grid(row=j+1, column=0, sticky=NW)

        self.cue_light_settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="DEFINE CUE LIGHTS", icon_name='light_bulb', icon_link=Links.CUE_LIGHTS.value)
        self.cue_light_cnt_dropdown = SimBADropDown(parent=self.cue_light_settings_frm, dropdown_options=list(range(1, len(self.roi_names)+1)), label='# CUE LIGHTS', label_width=20, dropdown_width=self.max_len+2, value=1, command= lambda x: _get_cue_light_names(int(x)))
        self.cue_light_settings_frm.grid(row=0, column=0, sticky=NW, padx=(0, 10))
        self.cue_light_cnt_dropdown.grid(row=0, column=0, sticky=NW)
        _get_cue_light_names()

        self.analyze_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="ANALYZE", icon_name='analyze_blue', icon_link=Links.CUE_LIGHTS.value)
        self.analyze_cue_light_data_btn = SimbaButton(parent=self.analyze_frm, txt="ANALYZE CUE LIGHT DATA", img='data_table', cmd=self._initialize_analysis_pop_up)
        self.visualize_cue_light_data_btn = SimbaButton(parent=self.analyze_frm, txt="VISUALIZE CUE LIGHT DATA", img='eye', cmd=self._initialize_visualize_pop_up)
        self.analyze_cue_light_movement_btn = SimbaButton(parent=self.analyze_frm, txt="ANALYZE CUE LIGHT MOVEMENT", img='analyze_green', cmd=self._initialize_movement_pop_up)
        self.analyze_cue_light_clf_btn = SimbaButton(parent=self.analyze_frm, txt="ANALYZE CUE LIGHT CLASSIFICATIONS", img='clf_2', cmd=self._initialize_clf_pop_up)
        lbl_info = SimBALabel(parent=self.analyze_frm, txt="[CLICK HERE TO LEARN ABOUT CUE LIGHT ANALYSIS]", link=Links.CUE_LIGHTS.value, cursor="hand2", txt_clr="blue")


        self.analyze_frm.grid(row=0, column=1, sticky=NW)
        lbl_info.grid(row=0, column=0, sticky=NW)
        self.analyze_cue_light_data_btn.grid(row=1, column=0, sticky=W)
        self.visualize_cue_light_data_btn.grid(row=2, column=0, sticky=W)
        self.analyze_cue_light_movement_btn.grid(row=3, column=0, sticky=W)
        self.analyze_cue_light_clf_btn.grid(row=4, column=0, sticky=W)

        #self.main_frm.mainloop()

    def __get_cue_light_names(self):
        return [x.get_value() for x in self.cue_light_dropdowns]

    def _initialize_analysis_pop_up(self):
        cue_light_names = self.__get_cue_light_names()
        _ = CueLightDataAnalyzerPopUp(config_path=self.config_path, cue_light_names=cue_light_names, data_dir=self.outlier_corrected_dir)

    def _initialize_visualize_pop_up(self):
        cue_light_names = self.__get_cue_light_names()
        _ = CueLightVisulizerPopUp(config_path=self.config_path, cue_light_names=cue_light_names)

    def _initialize_movement_pop_up(self):
        cue_light_names = self.__get_cue_light_names()
        _ = CueLightMovementAnalyzerPopUp(config_path=self.config_path, cue_light_names=cue_light_names)

    def _initialize_clf_pop_up(self):
        cue_light_names = self.__get_cue_light_names()
        _ = CueLightClfAnalyzerPopUp(config_path=self.config_path, cue_light_names=cue_light_names)



#est = CueLightMainPopUp(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini")



__author__ = "Simon Nilsson"

""" Tkinter pop-up classes for unsupervised ML"""

import os
import threading
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import DropDownMenu, FileSelect, FolderSelect
from simba.unsupervised.enums import UMLOptions
from simba.unsupervised.umap_embedder import UmapEmbedder
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists)
from simba.utils.enums import Formats


class TransformDimReductionPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, title="DIMENSIONALITY REDUCTION: TRANSFORM")
        ConfigReader.__init__(self, config_path=config_path)
        self.dim_reduction_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.model_select = FileSelect(
            self.dim_reduction_frm,
            "MODEL (PICKLE):",
            lblwidth=25,
            file_types=[("SimBA model pickle", f".{Formats.PICKLE.value}")],
            initialdir=self.project_path,
        )
        self.dataset_select = FileSelect(
            self.dim_reduction_frm,
            "DATASET (PICKLE):",
            lblwidth=25,
            file_types=[("SimBA model pickle", f".{Formats.PICKLE.value}")],
            initialdir=self.project_path,
        )
        self.save_dir = FolderSelect(
            self.dim_reduction_frm,
            "SAVE DIRECTORY: ",
            lblwidth=25,
            initialdir=self.project_path,
        )
        self.additional_data = DropDownMenu(
            self.dim_reduction_frm,
            "ADDITIONAL DATA:",
            UMLOptions.UMAP_ADDITIONAL_DATA.value,
            "25",
        )
        self.additional_data.setChoices("SCALED FEATURES & CLASSIFICATIONS")
        self.save_format_dropdown = DropDownMenu(
            self.dim_reduction_frm, "SAVE FORMATS:", UMLOptions.SAVE_FORMATS.value, "25"
        )
        self.save_format_dropdown.setChoices(UMLOptions.SAVE_FORMATS.value[0])

        self.dim_reduction_frm.grid(row=0, column=0, sticky=NW)
        self.model_select.grid(row=0, column=0, sticky=NW)
        self.dataset_select.grid(row=1, column=0, sticky=NW)
        self.save_dir.grid(row=2, column=0, sticky=NW)
        self.additional_data.grid(row=3, column=0, sticky=NW)
        self.save_format_dropdown.grid(row=4, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        check_if_dir_exists(in_dir=self.save_dir.folder_path)
        check_file_exist_and_readable(file_path=self.model_select.file_path)
        check_file_exist_and_readable(file_path=self.dataset_select.file_path)
        settings = {"DATA_FORMAT": None, "CLASSIFICATIONS": False}
        settings_choice = self.features_dropdown.getChoices()
        if settings_choice == "SCALED FEATURES":
            settings["DATA_FORMAT"] = "scaled"
        elif settings_choice == "RAW FEATURES":
            settings["DATA_FORMAT"] = "RAW"
        elif settings_choice == "SCALED FEATURES & CLASSIFICATIONS":
            settings["DATA_FORMAT"] = "scaled"
            settings["CLASSIFICATIONS"] = True
        elif settings_choice == "RAW FEATURES & CLASSIFICATIONS":
            settings["DATA_FORMAT"] = "RAW"
            settings["CLASSIFICATIONS"] = True

        umap_searcher = UmapEmbedder()
        threading.Thread(
            target=umap_searcher.transform(
                data_path=self.dataset_select.file_path,
                model=self.model_select.file_path,
                save_dir=self.save_dir.folder_path,
                settings=settings,
            )
        ).start()


# _ = TransformDimReductionPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/project_config.ini')




from tkinter import *

import numpy as np

from simba.data_processors.cuda.utils import _is_cuda_available
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.model.yolo_pose_inference import YOLOPoseInference
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FileSelect,
                                        FolderSelect, SimbaButton,
                                        SimBADropDown)
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists)
from simba.utils.enums import Options, PackageNames
from simba.utils.errors import SimBAGPUError, SimBAPAckageVersionError
from simba.utils.read_write import (find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    get_pkg_version, get_video_meta_data,
                                    str_2_bool)

MAX_TRACKS_OPTIONS = ['None', 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
BATCH_SIZE_OPTIONS =  list(range(50, 1050, 50))
CORE_CNT_OPTIONS = list(range(1, find_core_cnt()[0]))
IMG_SIZE_OPTIONS = [256, 320, 416, 480, 512, 640, 720, 768, 960, 1280]
devices = ['CPU']
THRESHOLD_OPTIONS = np.arange(0.1, 1.1, 0.1).astype(np.float32)


class YOLOPoseInferencePopUP(PopUpMixin):

    def __init__(self):
        gpu_available, gpus = _is_cuda_available()
        if not gpu_available:
            raise SimBAGPUError(msg=f'Cannot train YOLO pose-estimation model. No NVIDA GPUs detected on machine', source=self.__class__.__name__)
        ultralytics_version = get_pkg_version(pkg=PackageNames.ULTRALYTICS.value)
        if ultralytics_version is None:
            raise SimBAPAckageVersionError(msg=f'Cannot train YOLO pose-estimation model: Could not find ultralytics package in python environment',  source=self.__class__.__name__)

        PopUpMixin.__init__(self, title="PREDICT USING YOLO POSE ESTIMATION MODEL", icon='ultralytics_2')
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        devices.extend([f'{x} : {y["model"]}' for x, y in gpus.items()])
        self.weights_path = FileSelect(parent=settings_frm, fileDescription='MODEL PATH (E.G., .PT):', lblwidth=35,  entry_width=45, file_types=[("YOLO MODEL FILE", Options.ALL_YOLO_MODEL_FORMAT_STR_OPTIONS.value)])
        self.save_dir = FolderSelect(settings_frm, folderDescription="SAVE DIRECTORY:", lblwidth=35, entry_width=45)
        self.batch_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=BATCH_SIZE_OPTIONS, label="BATCH SIZE: ", label_width=35, dropdown_width=40, value=250)
        self.verbose_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="VERBOSE:", label_width=35, dropdown_width=40, value='TRUE')
        self.workers_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=CORE_CNT_OPTIONS, label="CPU WORKERS:", label_width=35, dropdown_width=40, value=int(max(CORE_CNT_OPTIONS)/2))
        self.format_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=Options.VALID_YOLO_FORMATS.value, label="FORMAT:", label_width=35, dropdown_width=40, value='engine')
        self.img_size_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=IMG_SIZE_OPTIONS, label="IMAGE SIZE:", label_width=35, dropdown_width=40, value=640)
        self.devices_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=devices, label="DEVICE:", label_width=35, dropdown_width=40, value=devices[1])
        self.interpolate_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="INTERPOLATE:",  label_width=35, dropdown_width=40, value='FALSE')
        self.stream_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="STREAM:", label_width=35, dropdown_width=40, value='FALSE')
        self.threshold_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=THRESHOLD_OPTIONS, label="THRESHOLD:", label_width=35, dropdown_width=40, value=0.7)
        self.iou_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=THRESHOLD_OPTIONS,  label="IOU:", label_width=35, dropdown_width=40, value=0.7)
        self.max_tracks_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=MAX_TRACKS_OPTIONS, label="MAX TRACKS:", label_width=35, dropdown_width=40, value=0.7)

        settings_frm.grid(row=0, column=0, sticky=NW)
        self.weights_path.grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=1, column=0, sticky=NW)
        self.format_dropdown.grid(row=2, column=0, sticky=NW)
        self.img_size_dropdown.grid(row=3, column=0, sticky=NW)
        self.threshold_dropdown.grid(row=4, column=0, sticky=NW)
        self.interpolate_dropdown.grid(row=5, column=0, sticky=NW)
        self.iou_dropdown.grid(row=6, column=0, sticky=NW)
        self.stream_dropdown.grid(row=7, column=0, sticky=NW)
        self.workers_dropdown.grid(row=8, column=0, sticky=NW)
        self.verbose_dropdown.grid(row=9, column=0, sticky=NW)
        self.devices_dropdown.grid(row=10, column=0, sticky=NW)
        self.max_tracks_dropdown.grid(row=11, column=0, sticky=NW)

        single_video_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="ANALYZE SINGLE VIDEO", icon_name='video')
        self.video_path = FileSelect(parent=single_video_frm, fileDescription='VIDEO PATH:', lblwidth=35, entry_width=45, file_types=[("VIDEO", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)])
        run_single_video_btn = SimbaButton(parent=single_video_frm, txt='ANALYZE SINGLE VIDEO', txt_clr='blue', img='rocket', cmd=self.run, cmd_kwargs={'multiple': False})

        single_video_frm.grid(row=1, column=0, sticky=NW)
        self.video_path.grid(row=0, column=0, sticky=NW)
        run_single_video_btn.grid(row=1, column=0, sticky=NW)


        video_dir_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="ANALYZE VIDEO DIRECTORY", icon_name='stack')
        self.video_dir = FolderSelect(parent=video_dir_frm, folderDescription='VIDEO DIRECTORY PATH:', entry_width=45, lblwidth=35)
        run_multiple_video_btn = SimbaButton(parent=video_dir_frm, txt='ANALYZE VIDEO DIRECTORY', txt_clr='blue', img='rocket', cmd=self.run, cmd_kwargs={'multiple': True})

        video_dir_frm.grid(row=2, column=0, sticky=NW)
        self.video_dir.grid(row=0, column=0, sticky=NW)
        run_multiple_video_btn.grid(row=1, column=0, sticky=NW)

        #self.main_frm.mainloop()

    def run(self, multiple: bool):
        mdl_path = self.weights_path.file_path
        save_dir = self.save_dir.folder_path
        format = self.format_dropdown.get_value()
        img_size = int(self.img_size_dropdown.get_value())
        batch_size = int(self.batch_dropdown.get_value())
        threshold = float(self.threshold_dropdown.get_value())
        interpolate = str_2_bool(self.interpolate_dropdown.get_value())
        iou = float(self.iou_dropdown.get_value())
        stream = str_2_bool(self.stream_dropdown.get_value())
        max_tracks = str_2_bool(self.stream_dropdown.get_value())
        verbose = str_2_bool(self.verbose_dropdown.get_value())
        workers = int(self.workers_dropdown.get_value())
        device = self.devices_dropdown.get_value()
        device = 'cpu' if device == 'CPU' else int(device.split(':', 1)[0])
        check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} SAVE DIRECTORY')
        check_file_exist_and_readable(file_path=mdl_path, raise_error=True)

        if multiple:
            check_if_dir_exists(in_dir=self.video_dir.folder_path, source=self.__class__.__name__, raise_error=True)
            video_paths = find_files_of_filetypes_in_directory(directory=self.video_dir.folder_path, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, raise_error=True)
        else:
            check_file_exist_and_readable(file_path=self.video_path.file_path, raise_error=True)
            video_paths = [self.video_path.file_path]
        for video_path in video_paths:
            _ = get_video_meta_data(video_path=video_path)

        runner = YOLOPoseInference(weights=mdl_path,
                                   video_path=video_paths,
                                   verbose=verbose,
                                   save_dir=save_dir,
                                   device=device, format=format,
                                   batch_size=batch_size,
                                   torch_threads=workers,
                                   box_threshold=threshold,
                                   max_tracks=max_tracks,
                                   interpolate=interpolate,
                                   imgsz=img_size,
                                   iou=iou,
                                   stream=stream)
        runner.run()



#YOLOPoseInferencePopUP()



__author__ = "Simon Nilsson"

import os

""" Tkinter pop-up classes for unsupervised ML"""

import glob
import threading
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import DropDownMenu, FileSelect
from simba.unsupervised.embedding_correlation_calculator import \
    EmbeddingCorrelationCalculator
from simba.unsupervised.enums import UMLOptions
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import Formats, Options


class EmbedderCorrelationsPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, title="EMBEDDING CORRELATIONS")
        ConfigReader.__init__(self, config_path=config_path)
        self.spearman_var = BooleanVar(value=True)
        self.pearsons_var = BooleanVar(value=True)
        self.kendall_var = BooleanVar(value=True)
        self.plots_var = BooleanVar(value=False)
        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.data_file_selected = FileSelect(
            self.data_frm,
            "DATASET (PICKLE):",
            initialdir=self.project_path,
            file_types=[("SimBA model (pickle", f"*.{Formats.PICKLE.value}")],
        )
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.data_file_selected.grid(row=0, column=0, sticky=NW)

        self.settings_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.spearman_cb = Checkbutton(
            self.settings_frm, text="SPEARMAN", variable=self.spearman_var
        )
        self.pearsons_cb = Checkbutton(
            self.settings_frm, text="PEARSONS", variable=self.pearsons_var
        )
        self.kendall_cb = Checkbutton(
            self.settings_frm, text="KENDALL", variable=self.kendall_var
        )
        self.plot_correlation_dropdown = DropDownMenu(
            self.settings_frm,
            "PLOT CORRELATION:",
            UMLOptions.CORRELATION_OPTIONS.value,
            "25",
        )
        self.plot_correlation_clr_dropdown = DropDownMenu(
            self.settings_frm, "PLOT PALETTE:", Options.PALETTE_OPTIONS.value, "25"
        )
        self.plot_correlation_dropdown.setChoices(
            UMLOptions.CORRELATION_OPTIONS.value[0]
        )
        self.plot_correlation_dropdown.disable()
        self.plot_correlation_clr_dropdown.setChoices(Options.PALETTE_OPTIONS.value[0])
        self.plot_correlation_clr_dropdown.disable()
        self.plots_cb = Checkbutton(
            self.settings_frm,
            text="PLOTS",
            variable=self.plots_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.plots_var,
                dropdown_menus=[
                    self.plot_correlation_dropdown,
                    self.plot_correlation_clr_dropdown,
                ],
            ),
        )

        self.settings_frm.grid(row=1, column=0, sticky=NW)
        self.spearman_cb.grid(row=0, column=0, sticky=NW)
        self.pearsons_cb.grid(row=1, column=0, sticky=NW)
        self.kendall_cb.grid(row=2, column=0, sticky=NW)
        self.plots_cb.grid(row=3, column=0, sticky=NW)
        self.plot_correlation_dropdown.grid(row=4, column=0, sticky=NW)
        self.plot_correlation_clr_dropdown.grid(row=5, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        check_file_exist_and_readable(self.data_file_selected.file_path)
        settings = {
            "correlations": [],
            "plots": {"create": False, "correlations": None, "palette": None},
        }
        if self.spearman_var.get():
            settings["correlations"].append("spearman")
        if self.pearsons_var.get():
            settings["correlations"].append("pearson")
        if self.kendall_var.get():
            settings["correlations"].append("kendall")
        if self.plots_var.get():
            settings["plots"]["create"] = True
            settings["plots"][
                "correlations"
            ] = self.plot_correlation_dropdown.getChoices()
            settings["plots"][
                "palette"
            ] = self.plot_correlation_clr_dropdown.getChoices()

        calculator = EmbeddingCorrelationCalculator(
            config_path=self.config_path,
            data_path=self.data_file_selected.file_path,
            settings=settings,
        )
        threading.Thread(target=calculator.run()).start()


# _ = EmbedderCorrelationsPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import os
from typing import Any, Dict, List, Optional, Union

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_if_keys_exist_in_dict, check_str,
    check_that_column_exist, check_valid_lst)
from simba.utils.enums import Formats
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df

STYLE_WIDTH = 'width'
STYLE_HEIGHT = 'height'
STYLE_FONT_SIZE = 'font size'
STYLE_LINE_WIDTH = 'line width'
STYLE_YMAX = 'y_max'
STYLE_COLOR = 'color'
STYLE_OPACITY = 'opacity'

AUTO = 'AUTO'

STYLE_ATTR = [STYLE_WIDTH, STYLE_HEIGHT, STYLE_FONT_SIZE, STYLE_LINE_WIDTH, STYLE_COLOR, STYLE_YMAX, STYLE_OPACITY]

class TresholdPlotCreatorSingleProcess(ConfigReader, PlottingMixin):
    """
    Create line chart visualizations displaying the classification probabilities of a single classifier.

    .. note::
       `Documentation <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`_.
       For improved run-time, use :meth:`simba.plotting.probability_plot_creator_mp.TresholdPlotCreatorMultiprocess`

    .. image:: _static/img/prob_plot.png
       :width: 300
       :align: center

    :param str config_path: path to SimBA project config file in Configparser format
    :param str clf_name: Name of the classifier to create visualizations for
    :param bool frame_setting: When True, SimBA creates indidvidual frames in png format
    :param bool video_setting: When True, SimBA creates compressed video in mp4 format
    :param bool last_image: When True, creates image .png representing last frame of the video.
    :param dict style_attr: User-defined style attributes of the visualization (line size, color etc).
    :param List[str] files_found: Files to create threshold plots for.

    Examples
    -----
    >>> style_attr = {'width': 640, 'height': 480, 'font size': 10, 'line width': 6, 'color': 'blue', 'circle size': 20}
    >>> clf_name='Attack'
    >>> files_found=['/_test/project_folder/csv/machine_results/Together_1.csv']

    >>> threshold_plot_creator = TresholdPlotCreatorSingleProcess(config_path='/_test/project_folder/project_config.ini', frame_setting=False, video_setting=True, last_frame=True, clf_name=clf_name, files_found=files_found, style_attr=style_attr)
    >>> threshold_plot_creator.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 files_found: List[Union[str, os.PathLike]],
                 clf_name: str,
                 style_attr: Dict[str, Any],
                 frame_setting: Optional[bool] = False,
                 video_setting: Optional[bool] = False,
                 last_image: Optional[bool] = True):

        if ((not frame_setting) and (not video_setting) and (not last_image)):
            raise NoSpecifiedOutputError(msg="Please choose to either probability videos, probability frames, and/or last frame.")

        check_file_exist_and_readable(file_path=config_path)
        check_valid_lst(data=files_found, source=self.__class__.__name__, valid_dtypes=(str,))
        print(style_attr)
        check_if_keys_exist_in_dict(data=style_attr, key=STYLE_ATTR, name=f'{self.__class__.__name__} style_attr')
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        check_str(name=f"{self.__class__.__name__} clf_name", value=clf_name, options=(self.clf_names))
        self.frame_setting, self.video_setting, self.style_attr, self.last_image = (frame_setting,video_setting,style_attr,last_image)
        self.files_found = files_found
        self.orginal_clf_name = clf_name
        self.clf_name = f"Probability_{self.orginal_clf_name}"
        self.out_width, self.out_height = (self.style_attr[STYLE_WIDTH], self.style_attr[STYLE_HEIGHT])
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        if not os.path.exists(self.probability_plot_dir):
            os.makedirs(self.probability_plot_dir)
        print(f"Processing {len(self.files_found)} video(s)...")
        self.timer = SimbaTimer(start=True)

    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.files_found)
        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            video_info, self.px_per_mm, fps = self.read_video_info(video_name=self.video_name)
            data_df = read_df(file_path, self.file_type)
            check_that_column_exist(df=data_df, column_name=self.clf_name, file_name=self.video_name)
            if self.frame_setting:
                self.save_frame_folder_dir = os.path.join(self.probability_plot_dir, f"{self.video_name}_{self.orginal_clf_name}")
                if not os.path.exists(self.save_frame_folder_dir):
                    os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                self.save_video_path = os.path.join(self.probability_plot_dir, f"{self.video_name}_{self.orginal_clf_name}.mp4")
                self.writer = cv2.VideoWriter(self.save_video_path, self.fourcc, fps, (self.out_width, self.out_height))


            clf_data = data_df[self.clf_name].values
            if self.style_attr[STYLE_YMAX] == AUTO:
                self.style_attr[STYLE_YMAX] = float(np.max(clf_data))
            if self.last_image:
                final_frm_save_path = os.path.join(self.probability_plot_dir, f'{self.video_name}_{self.orginal_clf_name}_final_frm_{self.datetime}.png')
                _ = PlottingMixin.make_line_plot(data=[clf_data],
                                                       colors=[self.style_attr[STYLE_COLOR]],
                                                       width=self.style_attr[STYLE_WIDTH],
                                                       height=self.style_attr[STYLE_HEIGHT],
                                                       line_width=self.style_attr[STYLE_LINE_WIDTH],
                                                       font_size=self.style_attr[STYLE_FONT_SIZE],
                                                       y_lbl=f"{self.clf_name} probability",
                                                       y_max= self.style_attr[STYLE_YMAX],
                                                       x_lbl='frame count',
                                                       title=self.clf_name,
                                                       save_path=final_frm_save_path,
                                                       line_opacity=self.style_attr[STYLE_OPACITY])

            if self.video_setting or self.frame_setting:
                for i in range(1, clf_data.shape[0]):
                    frm_data = clf_data[0:i]
                    img = PlottingMixin.make_line_plot(data=[frm_data],
                                                            colors=[self.style_attr[STYLE_COLOR]],
                                                            width=self.style_attr[STYLE_WIDTH],
                                                            height=self.style_attr[STYLE_HEIGHT],
                                                            line_width=self.style_attr[STYLE_LINE_WIDTH],
                                                            font_size=self.style_attr[STYLE_FONT_SIZE],
                                                            line_opacity=self.style_attr[STYLE_OPACITY],
                                                            y_lbl=f"{self.clf_name} probability",
                                                            y_max=self.style_attr[STYLE_YMAX],
                                                            x_lbl='frame count',
                                                            title=self.clf_name,
                                                            save_path=None)
                    if self.frame_setting:
                        frame_save_path = os.path.join(self.save_frame_folder_dir, f"{i}.png")
                        cv2.imwrite(frame_save_path, img)
                    if self.video_setting:
                        self.writer.write(img.astype(np.uint8)[:, :, :3])
                    print(f"Probability frame: {i+1} / {len(data_df)}. Video: {self.video_name} (File {file_cnt + 1}/{len(self.files_found)})")
                if self.video_setting:
                    self.writer.release()
                video_timer.stop_timer()
                print(f"Probability plot for video {self.video_name} saved (elapsed time: {video_timer.elapsed_time_str}s)...")
        self.timer.stop_timer()
        stdout_success(msg=f"All probability visualizations created in {self.probability_plot_dir} directory", elapsed_time=self.timer.elapsed_time_str)


# test = TresholdPlotCreatorSingleProcess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                         frame_setting=False,
#                                         video_setting=True,
#                                         last_image=True,
#                                         clf_name='Attack',
#                                         files_found=['/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                         style_attr={'width': 640, 'height': 480, 'font size': 10, 'line width': 10, 'color': 'Orange', 'circle size': 20, 'y_max': 'auto'})
# test.run()

# test = TresholdPlotCreatorSingleProcess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                         frame_setting=False,
#                                         video_setting=True,
#                                         last_image=True,
#                                         clf_name='Attack',
#                                         files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                         style_attr={'width': 640, 'height': 480, 'font size': 10, 'line width': 6, 'color': 'blue', 'circle size': 20, 'y_max': 'auto'})
# test.create_plots()


#



__author__ = "Simon Nilsson"

import os
from typing import Dict, Union

import numpy as np
import pandas as pd
from scipy.stats import f_oneway, kruskal
from statsmodels.stats.libqsturng import psturng
from statsmodels.stats.multicomp import pairwise_tukeyhsd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_keys_exist_in_dict)
from simba.utils.enums import Methods
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_unique_values_in_iterable, read_pickle

FEATURE_NAME = "FEATURE NAME"
FEATURE_IMPORTANCE = "IMPORTANCE"
F_STATISTIC = "F-STATISTIC"
MEASURE = "MEASURE"
P_VALUE = "P-VALUE"
CLUSTER = "CLUSTER"
PAIRED = "cluster_paired"
CORRELATION_METHODS = "correlation_methods"
GINI_IMPORTANCE = "gini_importance"
TUKEY = "tukey_posthoc"
METHOD = "method"
TARGET = "TARGET"
PEARSON = "pearson"
ANOVA = "anova"
KENDALL = "kendall"
SHAP = "shap"
SCALED = "scaled"
PLOTS = "plots"
CREATE = "create"
SPEARMAN = "spearman"
KRUSKAL_WALLIS = "kruskal_wallis"
MEAN = "MEAN"
STDEV = "STANDARD DEVIATION"
PERMUTATION_IMPORTANCE = "permutation_importance"
DESCRIPTIVE_STATISTICS = "descriptive_statistics"
ANOVA_HEADERS = ["FEATURE NAME", "F-STATISTIC", "P-VALUE"]
KRUSKAL_HEADERS = ["FEATURE NAME", "KRUSKAL-WALLIS H STATISTIC", "P-VALUE"]


class ClusterFrequentistCalculator(UnsupervisedMixin, ConfigReader):
    """
    Class for computing frequentist statitics based on cluster assignment labels for explainability purposes.

    :param Union[str, os.PathLike] config_path: path to SimBA configparser.ConfigParser project_config.ini
    :param Union[str, os.PathLike] data_path: path to pickle holding unsupervised results in ``simba.unsupervised.data_map.yaml`` format.
    :param dict settings: Dict holding which statistical tests to use, with test name as keys and booleans as values.

    :example:
    >>> settings = {'scaled': True, 'ANOVA': True, 'tukey_posthoc': True, 'descriptive_statistics': True}
    >>> calculator = ClusterFrequentistCalculator(config_path='unsupervised/project_folder/project_config.ini', data_path='unsupervised/cluster_models/quizzical_rhodes.pickle', settings=settings)
    >>> calculator.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_path: Union[str, os.PathLike],
        settings: Dict[str, bool],
    ):

        check_file_exist_and_readable(file_path=data_path)
        check_file_exist_and_readable(file_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        self.settings = settings
        self.data = read_pickle(data_path=data_path)
        self.save_path = os.path.join(
            self.logs_path,
            f"cluster_descriptive_statistics_{self.data[Clustering.CLUSTER_MODEL.value][Unsupervised.HASHED_NAME.value]}_{self.datetime}.xlsx",
        )
        check_if_keys_exist_in_dict(
            data=self.data,
            key=[Clustering.CLUSTER_MODEL.value, Unsupervised.METHODS.value],
            name=data_path,
        )
        check_if_keys_exist_in_dict(
            data=settings,
            key=[SCALED, ANOVA, DESCRIPTIVE_STATISTICS, TUKEY],
            name="settings",
        )

    def run(self):
        self.x_data = self.data[Unsupervised.METHODS.value][
            Unsupervised.SCALED_DATA.value
        ]
        self.cluster_data = self.data[Clustering.CLUSTER_MODEL.value][
            Unsupervised.MODEL.value
        ].labels_
        if not self.settings[SCALED]:
            self.x_data = TrainModelMixin.scaler_inverse_transform(
                data=self.x_data,
                scaler=self.data[Unsupervised.METHODS.value][Unsupervised.SCALER.value],
            )
        self.x_y_df = pd.concat(
            [
                self.x_data,
                pd.DataFrame(
                    self.cluster_data, columns=[CLUSTER], index=self.x_data.index
                ),
            ],
            axis=1,
        )

        self.cluster_cnt = get_unique_values_in_iterable(
            data=self.cluster_data,
            name=self.data[Clustering.CLUSTER_MODEL.value][
                Unsupervised.HASHED_NAME.value
            ],
            min=2,
        )
        with pd.ExcelWriter(self.save_path, mode="w") as writer:
            pd.DataFrame().to_excel(writer, sheet_name=" ", index=True)
        if self.settings[ANOVA]:
            self.__one_way_anovas()
        if self.settings[DESCRIPTIVE_STATISTICS]:
            self.__descriptive_stats()
        if self.settings[TUKEY]:
            self.__tukey_posthoc()
        if self.settings[KRUSKAL_WALLIS]:
            self.__kruskal_wallis()

        self.timer.stop_timer()
        stdout_success(
            msg=f"Cluster statistics complete. Data saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __save_results(self, df: pd.DataFrame, name: str):
        with pd.ExcelWriter(self.save_path, mode="a") as writer:
            df.to_excel(writer, sheet_name=name, index=True)

    def __one_way_anovas(self):
        print("Calculating ANOVAs...")
        timer = SimbaTimer(start=True)
        self.anova_results = pd.DataFrame(columns=ANOVA_HEADERS)
        for feature_name in self.data[Unsupervised.METHODS.value][
            Unsupervised.FEATURE_NAMES.value
        ]:
            stats_data = (
                self.x_y_df[[feature_name, "CLUSTER"]]
                .sort_values(by=["CLUSTER"])
                .values
            )
            stats_data = np.split(
                stats_data[:, 0], np.unique(stats_data[:, 1], return_index=True)[1][1:]
            )
            f_val, p_val = f_oneway(*stats_data)
            self.anova_results.loc[len(self.anova_results)] = [
                feature_name,
                f_val,
                p_val,
            ]
        self.anova_results = self.anova_results.sort_values(by=[P_VALUE]).set_index(
            FEATURE_NAME
        )
        self.anova_results[P_VALUE] = self.anova_results[P_VALUE].round(5)
        self.__save_results(df=self.anova_results, name=Methods.ANOVA.value)
        timer.stop_timer()
        stdout_success(
            msg=f"ANOVAs saved in {self.save_path}", elapsed_time=timer.elapsed_time_str
        )

    def __descriptive_stats(self):
        print("Calculating descriptive statistics..")
        timer = SimbaTimer(start=True)
        self.descriptive_results = []
        for feature_name in self.data[Unsupervised.METHODS.value][
            Unsupervised.FEATURE_NAMES.value
        ]:
            agg = (
                self.x_y_df.groupby([CLUSTER])[feature_name]
                .agg(["mean", "std", "sem"])
                .T
            )
            agg[FEATURE_NAME] = feature_name
            agg = (
                agg.reset_index(drop=False)
                .set_index(FEATURE_NAME)
                .rename(columns={"index": MEASURE})
            )
            self.descriptive_results.append(pd.DataFrame(agg))
        self.descriptive_results = pd.concat(self.descriptive_results, axis=0)
        self.__save_results(df=self.descriptive_results, name=DESCRIPTIVE_STATISTICS)
        timer.stop_timer()
        stdout_success(
            msg=f"Descriptive statistics saved in {self.save_path}",
            elapsed_time=timer.elapsed_time_str,
        )

    def __tukey_posthoc(self):
        print("Calculating tukey posthocs...")
        timer = SimbaTimer(start=True)
        self.post_hoc_results = []
        for feature_name in self.data[Unsupervised.METHODS.value][
            Unsupervised.FEATURE_NAMES.value
        ]:
            data = pairwise_tukeyhsd(self.x_y_df[feature_name], self.x_y_df[CLUSTER])
            df = pd.DataFrame(
                data=data._results_table.data[1:], columns=data._results_table.data[0]
            )
            df[P_VALUE] = psturng(
                np.abs(data.meandiffs / data.std_pairs),
                len(data.groupsunique),
                data.df_total,
            )
            df[FEATURE_NAME] = feature_name
            df = df.reset_index(drop=True).set_index(FEATURE_NAME)
            self.post_hoc_results.append(df)
        self.post_hoc_results = pd.concat(self.post_hoc_results, axis=0)
        self.__save_results(df=self.post_hoc_results, name=TUKEY)
        timer.stop_timer()
        stdout_success(
            msg=f"Tukey post-hocs' statistics saved in {self.save_path}",
            elapsed_time=timer.elapsed_time_str,
        )

    def __kruskal_wallis(self):
        timer = SimbaTimer(start=True)
        print("Calculating Kruskal-Wallis...")
        results = pd.DataFrame(columns=KRUSKAL_HEADERS)
        for feature_name in self.data[Unsupervised.METHODS.value][
            Unsupervised.FEATURE_NAMES.value
        ]:
            feature_data = []
            for i in self.x_y_df[CLUSTER].unique():
                feature_data.append(
                    list(self.x_y_df[feature_name][self.x_y_df[CLUSTER] == i].values)
                )
            statistic, p_val = kruskal(*feature_data)
            results.loc[len(results)] = [feature_name, statistic, p_val]
        results = (
            results.reset_index(drop=True)
            .set_index(FEATURE_NAME)
            .sort_values("P-VALUE", ascending=True)
        )
        self.__save_results(df=results, name=KRUSKAL_WALLIS)
        timer.stop_timer()
        stdout_success(
            msg=f"Kruskal-Wallis statistics saved in {self.save_path}",
            elapsed_time=timer.elapsed_time_str,
        )

        # data = pairwise_tukeyhsd(self.x_y_df[feature_name], self.x_y_df[CLUSTER])


# settings = {'scaled': True,
#             'anova': False,
#             'tukey_posthoc': False,
#             'descriptive_statistics': False,
#             'kruskal_wallis': True}
# calculator = ClusterFrequentistCalculator(config_path='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini', data_path='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/small_clusters/adoring_hoover.pickle', settings=settings)
# calculator.run()



import functools
import multiprocessing
import os
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd
from shapely.geometry import (LineString, MultiLineString, MultiPolygon, Point,
                              Polygon)

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_float, check_if_dir_exists,
                                check_if_valid_rgb_tuple, check_instance,
                                check_int, check_iterable_length, check_str,
                                check_valid_boolean, check_valid_lst)
from simba.utils.data import create_color_palettes
from simba.utils.enums import Defaults, Formats
from simba.utils.errors import InvalidInputError
from simba.utils.lookups import get_color_dict
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt, find_video_of_file,
                                    get_fn_ext, get_video_meta_data)
from simba.utils.warnings import FrameRangeWarning

ACCEPTED_TYPES = [Polygon, LineString, MultiPolygon, MultiLineString, Point]
FRAME_COUNT = "frame_count"

def geometry_visualizer(data: Tuple[int, pd.DataFrame],
                        video_path: Union[str, os.PathLike],
                        video_temp_dir: Union[str, os.PathLike],
                        video_meta_data: dict,
                        thickness: int,
                        verbose: bool,
                        bg_opacity: float,
                        colors: list,
                        circle_size: int,
                        shape_opacity: float):

    group, idx = int(data[0]), data[1].index.tolist()
    start_frm, end_frm = idx[0], idx[-1]
    fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
    video_save_path = os.path.join(video_temp_dir, f"{group}.mp4")
    video_writer = cv2.VideoWriter(video_save_path, fourcc, video_meta_data["fps"], (video_meta_data["width"], video_meta_data["height"]))
    cap = cv2.VideoCapture(video_path)
    batch_shapes = data[1].values.reshape(len(data[1]), -1)
    for frm_cnt, frm_id in enumerate(range(start_frm, end_frm + 1)):
        cap.set(1, int(frm_id))
        ret, img = cap.read()
        if ret:
            img_cpy = img.copy()
            if bg_opacity != 1.0:
                opacity = 1 - bg_opacity
                h, w, clr = img.shape[:3]
                opacity_image = np.ones((h, w, clr), dtype=np.uint8) * int(255 * opacity)
                img = cv2.addWeighted(img.astype(np.uint8), 1 - opacity, opacity_image.astype(np.uint8), opacity, 0)
            for shape_cnt, shape in enumerate(batch_shapes[frm_cnt]):
                if isinstance(shape, Polygon):
                    img_cpy = cv2.fillPoly(img_cpy, [np.array(shape.exterior.coords).astype(np.int32)], color=(colors[shape_cnt]))
                    interior_coords = [np.array(interior.coords, dtype=np.int32).reshape((-1, 1, 2)) for interior in shape.interiors]
                    for interior in interior_coords:
                        img_cpy = cv2.fillPoly(img_cpy, [interior], color=(colors[shape_cnt][::-1]))
                elif isinstance(shape, LineString):
                    img_cpy = cv2.fillPoly(img_cpy, [np.array(shape.coords, dtype=np.int32)], color=(colors[shape_cnt]))
                elif isinstance(shape, MultiPolygon):
                    for polygon_cnt, polygon in enumerate(shape.geoms):
                        img_cpy = cv2.fillPoly(img_cpy, [np.array((polygon.convex_hull.exterior.coords), dtype=np.int32)], color=(colors[shape_cnt]))
                elif isinstance(shape, MultiLineString):
                    for line_cnt, line in enumerate(shape.geoms):
                        img_cpy = cv2.fillPoly(img_cpy,[np.array(shape[line_cnt].coords, dtype=np.int32)], color=(colors[shape_cnt]))
                elif isinstance(shape, Point):
                    arr = np.array((shape.coords)).astype(np.int32)
                    x, y = arr[0][0], arr[0][1]
                    img_cpy = cv2.circle(img_cpy,(x, y), circle_size, colors[shape_cnt], thickness)
            if shape_opacity is not None:
                img = cv2.addWeighted(img_cpy, shape_opacity, img, 1 - shape_opacity, 0, img)
            else:
                img = np.copy(img_cpy)
            video_writer.write(img.astype(np.uint8))
            if verbose:
                print(f"Creating frame {frm_id} / {video_meta_data['frame_count']} (CPU core: {group}, video name: {video_meta_data['video_name']})")
        else:
            FrameRangeWarning(msg=f'Frame {frm_id} in video {video_meta_data["video_name"]} could not be read.')
            pass
    video_writer.release()
    cap.release()

    return group


class GeometryPlotter(ConfigReader, PlottingMixin):
    """
    A class for creating overlay geometry visualization videos based on provided geometries and video name.

    .. seealso::
       To quickly create static geometries on a white background (useful for troubleshooting unexpected geometries), see :func:`simba.mixins.geometry_mixin.GeometryMixin.view_shapes`
       and :func:`:func:`simba.mixins.geometry_mixin.GeometryMixin.geometry_video`

    .. image:: _static/img/GeometryPlotter.gif
       :width: 600
       :align: center

    :param config_path: Union[str, os.PathLike]: Path to SimBA configuration file.
    :param geometries: List[List[Union[Polygon, LineString, MultiPolygon, MultiLineString, Point]]]: List of lists of geometries for each frame.
    :param video_name: str: Name of the input video.
    :param core_cnt: int, optional: Number of CPU cores to use for parallel processing. Defaults to -1 represnting all available cores
    :raises InvalidInputError: If the provided geometries contain invalid data types.
    :raises CountError: If the number of shapes in the geometries does not match the number of frames in the video.
    """

    def __init__(self,
                 geometries: List[List[Union[Polygon, LineString, MultiPolygon, MultiLineString, Point]]],
                 video_name: Union[str, os.PathLike],
                 config_path: Optional[Union[str, os.PathLike]] = None,
                 core_cnt: Optional[int] = -1,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 thickness: Optional[int] = None,
                 circle_size: Optional[int] = None,
                 bg_opacity: Optional[float] = 1,
                 shape_opacity: float = 0.3,
                 palette: Optional[str] = None,
                 colors: Optional[List[Union[str, Tuple[int, int, int]]]] = None,
                 verbose: Optional[bool] = True):

        PlottingMixin.__init__(self)
        check_instance(source=self.__class__.__name__, instance=geometries, accepted_types=list)
        check_iterable_length(source=self.__class__.__name__, val=len(geometries), min=1)
        if circle_size is not None:
            check_int(name="circle_size", value=circle_size, min_value=1)
        if thickness is not None:
            check_int(name="thickness", value=thickness, min_value=1)
        check_float(name="video_opacity", value=bg_opacity, min_value=0.0, max_value=1.0)
        check_float(name="shape_opacity", value=shape_opacity, min_value=0.0, max_value=1.0)
        check_valid_boolean(value=verbose, source='verbose', raise_error=True)
        self.color_dict = get_color_dict()
        check_int(name="CORE COUNT", value=core_cnt, min_value=-1, raise_error=True, unaccepted_vals=[0])
        self.core_cnt = core_cnt
        if core_cnt == -1 or core_cnt > find_core_cnt()[0]:
            self.core_cnt = find_core_cnt()[0]
        if config_path is not None:
            ConfigReader.__init__(self, config_path=config_path)
        if os.path.isfile(video_name):
            self.video_path = video_name
        else:
            if config_path is None:
                raise InvalidInputError(msg=f'When providing a non-path video name, pass config_path')
            self.video_path = find_video_of_file(video_dir=self.video_dir, filename=video_name, raise_error=True)
        video_name = get_fn_ext(filepath=self.video_path)[1]
        self.video_meta_data = get_video_meta_data(video_path=self.video_path)
        self.shape_opacity = shape_opacity
        if circle_size is None:
            circle_size = self.get_optimal_circle_size(frame_size=(self.video_meta_data['width'], self.video_meta_data['height']), circle_frame_ratio=50)
        if thickness is None:
            thickness = circle_size
        if palette is None:
            self.colors = []
            check_valid_lst(data=colors, source=f'{self.__class__.__name__} colors', valid_dtypes=(str, tuple), exact_len=len(geometries))
            for clr in colors:
                if isinstance(clr, str):
                    check_str(name=f'{self.__class__.__name__} colors', value=clr, options=list(self.color_dict.keys()))
                else:
                    check_if_valid_rgb_tuple(data=clr)
                self.colors.append(clr)
        else:
            colors = create_color_palettes(no_animals=1, map_size=len(geometries) + 1, cmaps=[palette])
            self.colors = [x for xs in colors for x in xs]
        for i in range(len(geometries)):
            if len(geometries[i]) != self.video_meta_data[FRAME_COUNT]:
                FrameRangeWarning(msg=f"Geometry {i+1} contains {len(geometries[i])} shapes but video has {self.video_meta_data[FRAME_COUNT]} frames")
        self.geometries, self.video_name, self.thickness = geometries, video_name, thickness
        if config_path is None:
            check_if_dir_exists(in_dir=save_dir)
            self.save_dir = save_dir
        else:
            self.save_dir = os.path.join(self.frames_output_dir, "geometry_visualization")
        if not os.path.isdir(self.save_dir): os.makedirs(self.save_dir)
        self.temp_dir = os.path.join(self.save_dir, video_name, "temp")
        if not os.path.isdir(self.temp_dir): os.makedirs(self.temp_dir)
        self.save_path = os.path.join(self.save_dir, f'{video_name}.mp4')
        self.verbose, self.bg_opacity, self.palette = verbose, bg_opacity, palette
        self.circles_size = circle_size

    def run(self):
        video_timer = SimbaTimer(start=True)
        data = pd.DataFrame(self.geometries).T
        data = np.array_split(data, self.core_cnt)
        data_splits = []
        for i in range(len(data)): data_splits.append((i, data[i]))
        del data

        with multiprocessing.Pool(self.core_cnt, maxtasksperchild=Defaults.LARGE_MAX_TASK_PER_CHILD.value) as pool:
            constants = functools.partial(geometry_visualizer,
                                          video_path=self.video_path,
                                          video_temp_dir=self.temp_dir,
                                          video_meta_data=self.video_meta_data,
                                          thickness=self.thickness,
                                          verbose=self.verbose,
                                          bg_opacity=self.bg_opacity,
                                          colors=self.colors,
                                          circle_size=self.circles_size,
                                          shape_opacity=self.shape_opacity)
            for cnt, result in enumerate(pool.imap(constants, data_splits, chunksize=1)):
                print(f"Section {result}/{len(data_splits)} complete...")
            pool.terminate()
            pool.join()

        print(f"Joining {self.video_name} geometry video...")
        concatenate_videos_in_folder(in_folder=self.temp_dir, save_path=self.save_path, remove_splits=True)
        video_timer.stop_timer()
        stdout_success(msg=f"Geometry video {self.save_path} complete!", elapsed_time=video_timer.elapsed_time_str, source=self.__class__.__name__)



# if __name__ == '__main__':
#     from simba.utils.read_write import read_pickle
#     VIDEO_PATH = r"D:\ares\data\termite_2\termite.mp4"
#     DATA_PATH = r"D:\ares\data\termite_2\termite_2_geometries.pickle"
#     data = read_pickle(data_path=DATA_PATH)
#
#     get_video_meta_data(video_path=VIDEO_PATH)
#
#     geos = []
#
#     for k, v in data.items():
#         geos.append(list(v.values()))
#         #geos.append([subdict[i] for subdict in data.values()])
#
#     plotter = GeometryPlotter(geometries=geos, video_name=VIDEO_PATH, save_dir=r"D:\ares\data\termite_2\video", palette='Set1', core_cnt=32, bg_opacity=1)
#     plotter.run()
#     #max_frm = 9000



from tkinter import *

import numpy as np

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.blob_visualizer import BlobVisualizer
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FolderSelect,
                                        SimBADropDown)
from simba.utils.checks import check_if_dir_exists
from simba.utils.lookups import get_color_dict
from simba.utils.read_write import find_core_cnt

OPACITY_OPTIONS = list(np.round(np.arange(0.1, 1.1, 0.1), 2))
CIRCLE_SIZES = list(np.round(np.arange(1, 21, 1), 2))
CIRCLE_SIZES.insert(0, 'AUTO')

class BlobVisualizerPopUp(PopUpMixin):
    def __init__(self):
        self.color_dict = get_color_dict()
        clr_names = list(self.color_dict.keys())
        clr_names.insert(0, 'NONE')
        PopUpMixin.__init__(self, title="VISUALIZE BLOB TRACKING", icon='bubble_pink')
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings', icon_link=None)
        self.core_cnt_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=list(range(1, find_core_cnt()[0] + 1)), label="CPU CORE COUNT:", label_width=30, dropdown_width=20, value=find_core_cnt()[0])
        self.shape_opacity_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=OPACITY_OPTIONS, label="SHAPE OPACITY:", label_width=30, dropdown_width=20, value=0.7)
        self.bg_opacity_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=OPACITY_OPTIONS, label="BACKGROUND OPACITY:", label_width=30, dropdown_width=20, value=1.0)
        self.keypoint_sizes_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=CIRCLE_SIZES, label="KEY-POINT SIZES:", label_width=30, dropdown_width=20, value=CIRCLE_SIZES[0])
        self.hull_clr_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=clr_names, label="HULL COLOR:", label_width=30, dropdown_width=20, value='Pink')
        self.anterior_clr_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=clr_names, label="ANTERIOR COLOR:", label_width=30, dropdown_width=20, value='Green')
        self.posterior_clr_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=clr_names, label="POSTERIOR COLOR:", label_width=30, dropdown_width=20, value='Orange')
        self.center_clr_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=clr_names, label="CENTER COLOR:", label_width=30, dropdown_width=20, value='Cyan')
        self.left_clr_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=clr_names, label="LEFT COLOR:", label_width=30, dropdown_width=20, value='NONE')
        self.right_clr_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=clr_names, label="RIGHT COLOR:", label_width=30, dropdown_width=20, value='NONE')


        settings_frm.grid(row=0, column=0, sticky=NW)
        self.core_cnt_dropdown.grid(row=0, column=0, sticky=NW)
        self.shape_opacity_dropdown.grid(row=1, column=0, sticky=NW)
        self.bg_opacity_dropdown.grid(row=2, column=0, sticky=NW)
        self.keypoint_sizes_dropdown.grid(row=3, column=0, sticky=NW)
        self.hull_clr_dropdown.grid(row=4, column=0, sticky=NW)
        self.anterior_clr_dropdown.grid(row=5, column=0, sticky=NW)
        self.posterior_clr_dropdown.grid(row=6, column=0, sticky=NW)
        self.center_clr_dropdown.grid(row=7, column=0, sticky=NW)
        self.left_clr_dropdown.grid(row=8, column=0, sticky=NW)
        self.right_clr_dropdown.grid(row=9, column=0, sticky=NW)

        dir_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="DIRECTORIES", icon_name='browse', icon_link=None)
        self.videos_dir = FolderSelect(parent=dir_frm, folderDescription='VIDEO DIRECTORY:', lblwidth=30, entry_width=20)
        self.data_dir = FolderSelect(parent=dir_frm, folderDescription='DATA DIRECTORY:', lblwidth=30, entry_width=20)
        self.save_dir = FolderSelect(parent=dir_frm, folderDescription='SAVE DIRECTORY:', lblwidth=30, entry_width=20)

        dir_frm.grid(row=1, column=0, sticky=NW)
        self.videos_dir.grid(row=0, column=0, sticky=NW)
        self.data_dir.grid(row=1, column=0, sticky=NW)
        self.save_dir.grid(row=2, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        core_cnt = int(self.core_cnt_dropdown.get_value())
        shape_opacity = float(self.shape_opacity_dropdown.get_value())
        bg_opacity = float(self.bg_opacity_dropdown.get_value())
        videos_dir = self.videos_dir.folder_path
        data_dir = self.data_dir.folder_path
        save_dir = self.save_dir.folder_path

        circle_size = None if self.keypoint_sizes_dropdown.get_value() == 'AUTO' else int(self.keypoint_sizes_dropdown.get_value())
        hull_clr = None if self.hull_clr_dropdown.get_value() == 'NONE' else self.colors_dict[self.hull_clr_dropdown.get_value()]
        anterior_clr = None if self.anterior_clr_dropdown.get_value() == 'NONE' else self.colors_dict[self.anterior_clr_dropdown.get_value()]
        posterior_clr = None if self.posterior_clr_dropdown.get_value() == 'NONE' else self.colors_dict[self.posterior_clr_dropdown.get_value()]
        center_clr = None if self.center_clr_dropdown.get_value() == 'NONE' else self.colors_dict[self.center_clr_dropdown.get_value()]
        left_clr = None if self.left_clr_dropdown.get_value() == 'NONE' else self.colors_dict[self.left_clr_dropdown.get_value()]
        right_clr = None if self.right_clr_dropdown.get_value() == 'NONE' else self.colors_dict[self.right_clr_dropdown.get_value()]

        check_if_dir_exists(in_dir=videos_dir, source=self.__class__.__name__)
        check_if_dir_exists(in_dir=data_dir, source=self.__class__.__name__)
        check_if_dir_exists(in_dir=save_dir, source=self.__class__.__name__)

        visualizer = BlobVisualizer(data_path=data_dir,
                                    video_path=videos_dir,
                                    save_dir=save_dir,
                                    core_cnt=core_cnt,
                                    shape_opacity=shape_opacity,
                                    bg_opacity=bg_opacity,
                                    circle_size=circle_size,
                                    hull=hull_clr,
                                    anterior=anterior_clr,
                                    posterior=posterior_clr,
                                    center=center_clr,
                                    left=left_clr,
                                    right=right_clr)

        visualizer.run()

#BlobVisualizerPopUp()


__author__ = "Simon Nilsson"

import os
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.feature_extraction_supplement_mixin import \
    FeatureExtractionSupplemental
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_float, check_that_column_exist,
    check_valid_lst)
from simba.utils.data import detect_bouts, slice_roi_dict_for_video
from simba.utils.enums import Keys
from simba.utils.errors import (CountError, MissingColumnsError,
                                ROICoordinatesNotFoundError)
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_data_paths, read_df
from simba.utils.warnings import NoDataFoundWarning


class ROIAnalyzer(ConfigReader, FeatureExtractionMixin):
    """
    Analyze movements, entries, exits, and time-spent-in user-defined ROIs. Results are stored in the
    'project_folder/logs' directory of the SimBA project.

    :param str config_path: Path to SimBA project config file in Configparser format.
    :param Optional[str] data_path: Path to folder or file holding the data used to calculate ROI aggregate statistics. If None, then defaults to the `project_folder/csv/outlier_corrected_movement_location` directory of the SimBA project. Default: None.
    :param Optional[bool] calculate_distances: If True, then calculate movements aggregate statistics (distances and velocities) inside ROIs. Results are saved in ``project_folder/logs/`` directory. Default: False.
    :param Optional[bool] detailed_bout_data: If True, saves a file with a row for every entry into each ROI for each animal in each video. Results are saved in ``project_folder/logs/`` directory. Default: False.
    :param Optional[float] threshold: Float between 0 and 1. Body-part locations detected below this confidence threshold are filtered. Default: 0.0.
    :param Optional[List[str]] body_parts: List of body-parts to perform ROI analysis on.

    .. note::
       `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

    :example:
    >>> test = ROIAnalyzer(config_path = r"/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini", calculate_distances=True, detailed_bout_data=True, body_parts=['Nose_1', 'Nose_2'], threshold=0.0)
    >>> test.run()
    >>> test.save()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_path: Optional[Union[str, os.PathLike, List[str]]] = None,
                 detailed_bout_data: bool = False,
                 calculate_distances: bool = False,
                 threshold: float = 0.0,
                 body_parts: Optional[List[str]] = None):

        check_file_exist_and_readable(file_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(expected_file_path=self.roi_coordinates_path)
        self.read_roi_data()
        FeatureExtractionMixin.__init__(self)
        self.data_paths = read_data_paths(path=data_path, default=self.outlier_corrected_paths, default_name=self.outlier_corrected_dir, file_type=self.file_type)
        check_float(name="Body-part probability threshold", value=threshold, min_value=0.0, max_value=1.0)
        check_valid_lst(data=body_parts, source=f"{self.__class__.__name__} body-parts", valid_dtypes=(str,))
        if len(set(body_parts)) != len(body_parts):
            raise CountError(msg=f"All body-part entries have to be unique. Got {body_parts}", source=self.__class__.__name__)
        self.bp_dict, self.bp_lk = {}, {}
        for bp in body_parts:
            animal = self.find_animal_name_from_body_part_name(bp_name=bp, bp_dict=self.animal_bp_dict)
            self.bp_dict[animal] = [f'{bp}_{"x"}', f'{bp}_{"y"}', f'{bp}_{"p"}']
            self.bp_lk[animal] = bp
        self.roi_headers = [v for k, v in self.bp_dict.items()]
        self.roi_headers = [item for sublist in self.roi_headers for item in sublist]
        self.calculate_distances, self.threshold = calculate_distances, threshold
        self.detailed_bout_data = detailed_bout_data

    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        self.movements_df = pd.DataFrame(columns=["VIDEO", "ANIMAL", "SHAPE", "MEASUREMENT", "VALUE"])
        self.entry_results = pd.DataFrame(columns=["VIDEO", "ANIMAL", "SHAPE", "ENTRY COUNT"])
        self.time_results = pd.DataFrame(columns=["VIDEO", "ANIMAL", "SHAPE", "TIME (S)"])
        self.roi_bout_results = []
        self.detailed_df = None
        for file_cnt, file_path in enumerate(self.data_paths):
            _, video_name, _ = get_fn_ext(file_path)
            print(f"Analysing ROI data for video {video_name}... (Video {file_cnt+1}/{len(self.data_paths)})")
            video_settings, pix_per_mm, self.fps = self.read_video_info(video_name=video_name)
            self.sliced_roi_dict, video_shape_names = slice_roi_dict_for_video(data=self.roi_dict, video_name=video_name)
            if len(video_shape_names) == 0:
                NoDataFoundWarning(msg=f"Skipping video {video_name}: No user-defined ROI data found for this video...")
                continue
            else:
                self.data_df = read_df(file_path, self.file_type).reset_index(drop=True)
                if len(self.bp_headers) != len(self.data_df.columns):
                    raise MissingColumnsError(msg=f"The data file {file_path} contains {len(self.data_df.columns)} body-part columns, but the project is made for {len(self.bp_headers)} body-part columns as suggested by the {self.body_parts_path} file", source=self.__class__.__name__)
                self.data_df.columns = self.bp_headers
                check_that_column_exist(df=self.data_df, column_name=self.roi_headers, file_name=file_path)
                for animal_name, bp_names in self.bp_dict.items():
                    animal_df = self.data_df[self.bp_dict[animal_name]].reset_index(drop=True)
                    animal_bout_results = {}
                    for _, row in self.sliced_roi_dict[Keys.ROI_RECTANGLES.value].iterrows():
                        roi_coords = np.array([[row["topLeftX"], row["topLeftY"]], [row["Bottom_right_X"], row["Bottom_right_Y"]]])
                        animal_df[row["Name"]] = (FeatureExtractionMixin.framewise_inside_rectangle_roi(bp_location=animal_df.values[:, 0:2], roi_coords=roi_coords))
                        animal_df.loc[animal_df[bp_names[2]] < self.threshold, row["Name"]] = 0
                        roi_bouts = detect_bouts(data_df=animal_df, target_lst=[row["Name"]], fps=self.fps)
                        roi_bouts["ANIMAL"] = animal_name
                        roi_bouts["VIDEO"] = video_name
                        self.roi_bout_results.append(roi_bouts)
                        animal_bout_results[row["Name"]] = roi_bouts
                        self.entry_results.loc[len(self.entry_results)] = [video_name,animal_name,row["Name"], len(roi_bouts)]
                        self.time_results.loc[len(self.time_results)] = [video_name,animal_name,row["Name"],roi_bouts["Bout_time"].sum()]
                    for _, row in self.sliced_roi_dict[Keys.ROI_CIRCLES.value].iterrows():

                        center_x, center_y = row["centerX"], row["centerY"]
                        animal_df[f'{row["Name"]}_distance'] = (FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=animal_df.values[:, 0:2], location_2=np.array([center_x, center_y]), px_per_mm=1))
                        animal_df[row["Name"]] = 0
                        animal_df.loc[animal_df[f'{row["Name"]}_distance'] <= row["radius"], row["Name"]] = 1
                        animal_df.loc[animal_df[bp_names[2]] < self.threshold, row["Name"]] = 0
                        roi_bouts = detect_bouts(data_df=animal_df, target_lst=[row["Name"]], fps=self.fps)
                        roi_bouts["ANIMAL"] = animal_name
                        roi_bouts["VIDEO"] = video_name
                        self.roi_bout_results.append(roi_bouts)
                        animal_bout_results[row["Name"]] = roi_bouts
                        self.entry_results.loc[len(self.entry_results)] = [video_name, animal_name, row["Name"], len(roi_bouts)]
                        self.time_results.loc[len(self.time_results)] = [video_name,animal_name,row["Name"],roi_bouts["Bout_time"].sum()]
                    for _, row in self.sliced_roi_dict[Keys.ROI_POLYGONS.value].iterrows():
                        roi_coords = np.array(list(zip(row["vertices"][:, 0], row["vertices"][:, 1])))
                        animal_df[row["Name"]] = (FeatureExtractionMixin.framewise_inside_polygon_roi(bp_location=animal_df.values[:, 0:2], roi_coords=roi_coords))
                        animal_df.loc[
                            animal_df[bp_names[2]] < self.threshold, row["Name"]
                        ] = 0
                        roi_bouts = detect_bouts(
                            data_df=animal_df, target_lst=[row["Name"]], fps=self.fps
                        )
                        roi_bouts["ANIMAL"] = animal_name
                        roi_bouts["VIDEO"] = video_name
                        self.roi_bout_results.append(roi_bouts)
                        animal_bout_results[row["Name"]] = roi_bouts
                        self.entry_results.loc[len(self.entry_results)] = [
                            video_name,
                            animal_name,
                            row["Name"],
                            len(roi_bouts),
                        ]
                        self.time_results.loc[len(self.time_results)] = [
                            video_name,
                            animal_name,
                            row["Name"],
                            roi_bouts["Bout_time"].sum(),
                        ]
                    if self.calculate_distances:
                        for roi_name, roi_data in animal_bout_results.items():
                            if len(roi_data) == 0:
                                self.movements_df.loc[len(self.movements_df)] = [video_name, animal_name, roi_name, "Movement (cm)", 0,]
                                self.movements_df.loc[len(self.movements_df)] = [video_name, animal_name, roi_name, "Average velocity (cm/s)", "None",]
                            else:
                                distances, velocities = [], []
                                roi_frames = roi_data[["Start_frame", "End_frame"]].values
                                for event in roi_frames:
                                    event_pose = animal_df.loc[np.arange(event[0], event[1] + 1), bp_names]
                                    event_pose = event_pose[event_pose[bp_names[2]] > self.threshold][bp_names[:2]].values
                                    if event_pose.shape[0] > 1:
                                        distance, velocity = (FeatureExtractionSupplemental.distance_and_velocity(x=event_pose, fps=self.fps, pixels_per_mm=pix_per_mm, centimeters=True))
                                        distances.append(distance)
                                        velocities.append(velocity)
                                self.movements_df.loc[len(self.movements_df)] = [video_name, animal_name, roi_name, "Movement (cm)", sum(distances)]
                                self.movements_df.loc[len(self.movements_df)] = [video_name, animal_name, roi_name, "Average velocity (cm/s)", np.average(velocities)]

        if len(self.roi_bout_results) > 0:
            self.detailed_df = pd.concat(self.roi_bout_results, axis=0)
            self.detailed_df = self.detailed_df.rename(columns={"Event": "SHAPE NAME", "Start_time": "START TIME", "End Time": "END TIME", "Start_frame": "START FRAME", "End_frame": "END FRAME", "Bout_time": "DURATION (S)"})
            self.detailed_df["BODY-PART"] = self.detailed_df["ANIMAL"].map(self.bp_lk)
            self.detailed_df = self.detailed_df[["VIDEO", "ANIMAL", "BODY-PART", "SHAPE NAME", "START TIME", "END TIME", "START FRAME", "END FRAME", "DURATION (S)"]].reset_index(drop=True)

    def save(self):
        self.entry_results["BODY-PART"] = self.entry_results["ANIMAL"].map(self.bp_lk)
        self.time_results["BODY-PART"] = self.time_results["ANIMAL"].map(self.bp_lk)
        self.entry_results = self.entry_results[["VIDEO", "ANIMAL", "BODY-PART", "SHAPE", "ENTRY COUNT"]]
        self.time_results = self.time_results[["VIDEO", "ANIMAL", "BODY-PART", "SHAPE", "TIME (S)"]]
        self.entry_results.to_csv(os.path.join(self.logs_path, f'{"ROI_entry_data"}_{self.datetime}.csv'))
        self.time_results.to_csv(os.path.join(self.logs_path, f'{"ROI_time_data"}_{self.datetime}.csv'))
        if self.detailed_bout_data and self.detailed_df is not None:
            detailed_path = os.path.join(self.logs_path, f'{"Detailed_ROI_data"}_{self.datetime}.csv')
            self.detailed_df.to_csv(detailed_path)
            print(f"Detailed ROI data saved at {detailed_path}...")
        if self.calculate_distances:
            movement_path = os.path.join(self.logs_path, f'{"ROI_movement_data"}_{self.datetime}.csv')
            self.movements_df["BODY-PART"] = self.movements_df["ANIMAL"].map(self.bp_lk)
            self.movements_df = self.movements_df[["VIDEO", "ANIMAL", "BODY-PART", "SHAPE", "MEASUREMENT", "VALUE"]]
            self.movements_df.to_csv(movement_path)
            print(f"ROI aggregate movement data saved at {movement_path}...")
        stdout_success(msg=f"ROI time and ROI entry saved in the {self.logs_path} directory in CSV format.")


# test = ROIAnalyzer(config_path = r"C:\troubleshooting\ROI_movement_test\project_folder\project_config.ini",
#                    data_path=None,
#                    calculate_distances=True,
#                    detailed_bout_data=True,
#                    body_parts=['Head'],
#                    threshold=0.0)
# test.run()




# test = ROIAnalyzer(config_path = r"/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini",
#                    data_path=None,
#                    calculate_distances=True,
#                    detailed_bout_data=True,
#                    body_parts=['Nose_1', 'Nose_2'],
#                    threshold=0.0)
# test.run()
# test.save()


#
# test = ROIAnalyzer(ini_path = r"/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini",
#                    data_path = "outlier_corrected_movement_location",
#                    calculate_distances=True,
#                    settings={'threshold': 0.00, 'body_parts': {'Animal_1': 'Nose_1'}})
# test.run()
# test.save()

# test = ROIAnalyzer(ini_path = r"/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini",
#                    data_path = "outlier_corrected_movement_location",
#                    calculate_distances=True)
# test.run()

# test = ROIAnalyzer(ini_path = r"/Users/simon/Desktop/envs/simba_dev/tests/data/test_projects/zebrafish/project_folder/project_config.ini",
#                    data_path = "outlier_corrected_movement_location",
#                    calculate_distances=True)


# settings = {'body_parts': {'animal_1_bp': 'Ear_left_1', 'animal_2_bp': 'Ear_left_2', 'animal_3_bp': 'Ear_right_1',}, 'threshold': 0.4}
# test = ROIAnalyzer(ini_path = r"/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini",
#                    data_path = "outlier_corrected_movement_location",
#                    settings=settings,
#                    calculate_distances=True)
# test.run()
# test.save()


# settings = {'body_parts': {'Simon': 'Ear_left_1', 'JJ': 'Ear_left_2'}, 'threshold': 0.4}
# test = ROIAnalyzer(ini_path = r"/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini",
#                    data_path = "outlier_corrected_movement_location",
#                    settings=settings,
#                    calculate_distances=True)
# test.read_roi_dfs()
# test.analyze_ROIs()
# test.save_data()


# settings = {'body_parts': {'animal_1_bp': 'Ear_left_1', 'animal_2_bp': 'Ear_left_2'}, 'threshold': 0.4}
# test = ROIAnalyzer(ini_path = r"/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini",
#                    data_path = "outlier_corrected_movement_location",
#                    calculate_distances=True)
# test.run()
# test.analyze_ROIs()
# test.save_data()



__author__ = "Simon Nilsson"
__email__ = "sronilsson@gmail.com"

import os
from typing import List, Optional, Tuple, Union

import numpy as np

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import functools
import multiprocessing

from numba import jit

from simba.data_processors.blob_location_computer import BlobLocationComputer
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_if_dir_exists, check_if_valid_rgb_tuple,
                                check_int, check_nvidea_gpu_available,
                                check_str, check_valid_boolean)
from simba.utils.enums import Defaults, Options, TextOptions
from simba.utils.errors import FFMPEGCodecGPUError, InvalidInputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_all_videos_in_directory,
                                    find_core_cnt, get_fn_ext,
                                    get_video_meta_data)


class BlobPlotter(PlottingMixin):
    """
     Plot the results of animal tracking based on blob.

     .. seealso::
        :func:`simba.mixins.plotting_mixin.PlottingMixin._plot_blobs`, :func:`simba.mixins.image_mixin.ImageMixin.get_blob_locations`

     .. video:: _static/img/BlobLocationComputer.webm
        :width: 800
        :autoplay:
        :loop:

     :param Union[List[str], str, os.PathLike] data_path: Path(s) to video file(s) or directory containing video files.
     :param Optional[bool] gpu: Whether to use GPU for processing. Defaults to False.
     :param Optional[int] batch_size: Number of frames to process in each batch. Defaults to 2000. Increase if your RAM allows.
     :param Optional[Tuple[int, int, int]] circle_color: Color of the blobs as an RGB tuple. Defaults to pink (255, 105, 180).
     :param Optional[Union[str, os.PathLike]] save_dir: Directory to save output files. If None, no files will be saved.
     :param Optional[int] verbose: If True, then prints msgs informing on progress.
     :param Optional[str] smoothing: Savitzky Golay, Gaussian, or None. Smooths body-part coordinate data for more accurate blob representation. Default None.
     :param Optional[int] circle_size: The circle defining the x, y location of the animal in the data. Defaults to None and SimBA will try and retrieve the optimal circle size based in the video resolution.
     :param Optional[int] core_cnt: The number of cores to use for multiprocessing. Deafults to -1 which means all available cores.

     :example:
     >>> BlobPlotter(data_path=r"C:\troubleshooting\RAT_NOR\project_folder\videos\test\2022-06-20_NOB_DOT_4_downsampled.mp4", smoothing='Savitzky Golay', circle_size=10).run()
     """

    def __init__(self,
                 data_path: Union[List[str], str, os.PathLike],
                 gpu: Optional[bool] = False,
                 batch_size: Optional[int] = 2000,
                 circle_color: Optional[Tuple[int, int, int]] = (TextOptions.COLOR.value),
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 verbose: Optional[int] = True,
                 smoothing: Optional[str] = None,
                 circle_size: Optional[int] = None,
                 core_cnt: Optional[int] = -1):



        PlottingMixin.__init__(self)
        if os.path.isdir(data_path):
            self.data_paths = find_all_videos_in_directory(directory=data_path, as_dict=True, raise_error=True).values()
        elif os.path.isfile(data_path):
            self.data_paths = [data_path]
        else:
            raise InvalidInputError(msg=f'{data_path} is not a valid directory or video file path or directory path.')
        self.video_meta_data = {}
        for i in self.data_paths:
            self.video_meta_data[get_fn_ext(filepath=i)[1]] = get_video_meta_data(video_path=i)
        if save_dir is not None:
            check_if_dir_exists(in_dir=os.path.dirname(save_dir), source=self.__class__.__name__)
        check_valid_boolean(value=[verbose, gpu])
        if gpu and not check_nvidea_gpu_available():
            raise FFMPEGCodecGPUError(msg='No GPU detected.', source=self.__class__.__name__)
        if smoothing is not None:
            check_str(name=f'{self.__class__.__name__} smoothing', value=smoothing, options=Options.SMOOTHING_OPTIONS.value)
        check_int(name=f'{self.__class__.__name__} batch_size', value=batch_size, min_value=1)
        check_int(name=f'{self.__class__.__name__} core_cnt', value=core_cnt, min_value=-1)
        if (core_cnt == -1) or core_cnt > find_core_cnt()[0]:
            core_cnt = find_core_cnt()[0]
        check_if_valid_rgb_tuple(data=circle_color)
        self.verbose = verbose
        self.gpu = gpu
        self.batch_size = batch_size
        self.save_dir = save_dir
        self.smoothing = smoothing
        self.core_cnt = core_cnt
        self.circle_size = circle_size
        self.circle_color = circle_color

    @staticmethod
    @jit(nopython=True)
    def __insert_group_idx_column(data: np.array, group: int):
        group_col = np.full((data.shape[0], 1), group)
        return np.hstack((group_col, data))

    def run(self):
        timer = SimbaTimer(start=True)
        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(filepath=file_path)
            temp_dir = os.path.join(os.path.dirname(file_path), video_name, '_temp')
            check_if_dir_exists(in_dir=temp_dir, source=self.__class__.__name__, create_if_not_exist=True)
            if self.save_dir is None:
                save_path = os.path.join(os.path.dirname(file_path), video_name + '_blob_tracking.mp4')
            else:
                save_path = os.path.join(self.save_dir, video_name + '_blob_tracking.mp4')
            blob_locator = BlobLocationComputer(data_path=file_path, verbose=self.verbose, gpu=self.gpu, batch_size=self.batch_size, save_dir=None, smoothing=self.smoothing, multiprocessing=True)
            blob_locator.run()
            print(f'Creating blob location video for {video_name}...')
            if not os.path.isdir(temp_dir): os.makedirs(temp_dir)
            x_y = blob_locator.location_data[video_name]
            del blob_locator
            x_y = np.hstack([np.arange(0, x_y.shape[0]).reshape(-1, 1), x_y])
            x_y = np.array_split(x_y, self.core_cnt)
            bps = [self.__insert_group_idx_column(data=i, group=cnt) for cnt, i in enumerate(x_y)]
            if self.circle_size is None:
                self.circle_size = self.get_optimal_circle_size(frame_size=(int(self.video_meta_data[video_name]['width']), int(self.video_meta_data[video_name]['height'])), circle_frame_ratio=100)
            with multiprocessing.Pool(self.core_cnt, maxtasksperchild=Defaults.MAXIMUM_MAX_TASK_PER_CHILD.value) as pool:
                constants = functools.partial(PlottingMixin._plot_blobs, verbose=self.verbose, circle_size=self.circle_size, circle_clr=self.circle_color, video_path=file_path, temp_dir=temp_dir)
                for cnt, result in enumerate(pool.imap(constants, bps, chunksize=1)):
                    print(f'Completed section {result+1}...')
            concatenate_videos_in_folder(in_folder=temp_dir, save_path=save_path, video_format="avi", remove_splits=True)
            video_timer.stop_timer()
            stdout_success(msg=f'Blob video {video_name} complete and saved at {save_path}', elapsed_time=video_timer.elapsed_time_str)
        timer.stop_timer()
        stdout_success(msg=f'{len(self.data_paths)} blob videos saved.', elapsed_time=video_timer.elapsed_time_str)


#BlobPlotter(data_path=r"C:\troubleshooting\RAT_NOR\project_folder\videos\test\2022-06-20_NOB_DOT_4_downsampled.mp4", smoothing='Savitzky Golay', circle_size=30, gpu=True).run()


import os
from typing import Optional, Union

import cv2
import numpy as np

from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_float, check_if_dir_exists, check_str
from simba.utils.data import create_color_palettes
from simba.utils.enums import Formats, Options
from simba.utils.read_write import read_img, recursive_file_search


class Yolo2Imgs():


    def __init__(self,
                 yolo_dir: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 palette: Optional[str] = None,
                 circle_size: Optional[Union[float, int]] = None):

        check_if_dir_exists(in_dir=yolo_dir, source=f'{self.__class__.__name__} yolo_dir', raise_error=True)
        check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} save_dir', raise_error=True)
        pose_palettes = Options.PALETTE_OPTIONS_CATEGORICAL.value + Options.PALETTE_OPTIONS.value
        palette = 'Set1' if palette is None else palette
        check_str(name=f'{self.__class__.__name__} palette', value=palette, options=pose_palettes)
        if circle_size is not None: check_float(name=f'{self.__class__.__name__} circle_size', value=circle_size, min_value=0.1)
        self.yolo_dir, self.save_dir, self.palette = yolo_dir, save_dir, palette
        self.circle_size = circle_size
    def run(self):
        lbl_paths = recursive_file_search(directory=self.yolo_dir, extensions=['.txt'], as_dict=True, raise_error=True)
        img_paths = recursive_file_search(directory=self.yolo_dir, extensions=Options.ALL_IMAGE_FORMAT_OPTIONS.value, as_dict=True, raise_error=True)

        for lbl_cnt, (lbl_name, lbl_path) in enumerate(lbl_paths.items()):
            img = read_img(img_path=img_paths[lbl_name])
            h, w = img.shape[:2]
            img_circle = self.circle_size if self.circle_size is not None else PlottingMixin().get_optimal_circle_size(frame_size=(w, h), circle_frame_ratio=100)
            with open(lbl_path, "r") as file: lbls = file.read().strip().split('\n')
            lbls = [x.strip().split() for x in lbls if x.strip()]
            for obs in lbls:
                if len(obs) < 5:
                    continue
                obs_lbl = [x for x in obs if x != '']
                obs_bbox = [float(x) for x in obs_lbl[1:5]]
                obs_kp = [float(x) for x in obs_lbl[5:]]
                if len(obs_kp) % 3 != 0:
                    continue
                obs_kp = [obs_kp[i:i + 3] for i in range(0, len(obs_kp), 3)]

                # Convert bbox from normalized to pixel
                box_cx, box_cy = int(obs_bbox[0] * w), int(obs_bbox[1] * h)
                box_w, box_h = int(obs_bbox[2] * w), int(obs_bbox[3] * h)
                tl = (int(box_cx - box_w / 2), int(box_cy - box_h / 2))
                br = (int(box_cx + box_w / 2), int(box_cy + box_h / 2))
                color_lst = create_color_palettes(1, len(obs_kp) + 1)[0]

                img = cv2.rectangle(img, tl, br, tuple(color_lst[0]), 2, lineType=-1)
                # for kp_cnt, kp in enumerate(obs_kp):
                #     if kp[2] != 0:
                #         x, y = int(kp[0] * w), int(kp[1] * h)
                #         img = cv2.circle(img, center=(x, y), radius=img_circle, color=color_lst[kp_cnt+1], thickness=-1)
                cv2.imshow('asdasdasd', img)
                cv2.waitKey(300,)
                print(tl, br, img.shape, lbl_path)

            #break




                #

                #
                #
                #
                # obs_bbox = [int(obs_bbox[2] * w), int(obs_bbox[3] * h), int(obs_bbox[0] * w), int(obs_bbox[1] * h), ]

                #print(tl, br, img.shape)


                # obs_kp = [[int(x[0] * w), int(x[1] * w), int(x[2])] for x in obs_kp]
                # tl = (int(obs_bbox[0] - obs_bbox[2]), int(obs_bbox[1] - obs_bbox[3]))
                # br = (int(obs_bbox[0] + obs_bbox[2]), int(obs_bbox[1] + obs_bbox[3]))
                #

                #bbox = np.vstack([tl, tr, br, bl])

                #break
            #break
                #
                #





            #break





# runner = Yolo2Imgs(yolo_dir=r"D:\cvat_annotations\yolo_07032025\bbox_test", save_dir=r'D:\cvat_annotations\yolo_07032025\imgs')
# runner.run()


__author__ = "Simon Nilsson"

import threading
from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.third_party_label_appenders.converters import labelme_to_df
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FolderSelect,
                                        SimBADropDown)
from simba.utils.checks import check_if_dir_exists
from simba.utils.read_write import str_2_bool


class LabelmeDirectory2CSVPopUp(PopUpMixin):

    """
    :example:
    >>> LabelmeDirectory2CSVPopUp()
    """
    def __init__(self):
        super().__init__(title="LABELME DIRECTORY TO CSV")
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        self.labelme_dir = FolderSelect(settings_frm, folderDescription="LABELME DIRECTORY:", lblwidth=35)
        self.save_dir = FolderSelect(settings_frm, folderDescription="SAVE DIRECTORY:", lblwidth=35)
        self.greyscale_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="GREYSCALE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.pad_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="PAD IMAGES: ", label_width=35, dropdown_width=40, value='FALSE')
        self.size_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['FALSE', 'MINIMUM', 'MAXIMUM'], label="RESIZE IMAGES: ", label_width=35, dropdown_width=40, value='FALSE')
        self.normalize_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'],  label="NORMALIZE IMAGES: ", label_width=35, dropdown_width=40, value='FALSE')
        self.verbose_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'],  label="VERBOSE: ", label_width=35, dropdown_width=40, value='TRUE')

        settings_frm.grid(row=0, column=0, sticky=NW)
        self.labelme_dir.grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=1, column=0, sticky=NW)
        self.greyscale_dropdown.grid(row=2, column=0, sticky=NW)
        self.pad_dropdown.grid(row=3, column=0, sticky=NW)
        self.size_dropdown.grid(row=4, column=0, sticky=NW)
        self.normalize_dropdown.grid(row=5, column=0, sticky=NW)
        self.verbose_dropdown.grid(row=6, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        labelme_dir = self.labelme_dir.folder_path
        save_path = self.save_dir.folder_path
        greyscale = str_2_bool(self.greyscale_dropdown.get_value())
        pad = str_2_bool(self.pad_dropdown.get_value())
        normalize = str_2_bool(self.normalize_dropdown.get_value())
        verbose = str_2_bool(self.verbose_dropdown.get_value())

        size = self.size_dropdown.get_value()
        size = None if size == 'FALSE' else 'min' if size == 'MINIMUM' else 'max' if size == 'MAXIMUM' else size
        check_if_dir_exists(in_dir=labelme_dir, source=f'{self.__class__.__name__} LABELME DIRECTORY')
        check_if_dir_exists(in_dir=labelme_dir, source=f'{self.__class__.__name__} SAVE DIRECTORY')
        thread = threading.Thread(target=labelme_to_df, kwargs={'labelme_dir': labelme_dir, 'save_path': save_path, 'greyscale': greyscale, 'pad': pad, 'normalize': normalize, 'size': size, 'verbose': verbose})
        thread.start()


__author__ = "Simon Nilsson"

import os
from typing import Optional, Union

import numpy as np
import pandas as pd
from numba import jit, prange

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.data import slice_roi_dict_for_video
from simba.utils.enums import ROI_SETTINGS
from simba.utils.errors import (InvalidInputError, NoDataError,
                                ROICoordinatesNotFoundError)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_data_paths, read_df


class DirectingROIAnalyzer(ConfigReader, FeatureExtractionMixin):
    """
    Compute aggregate statistics for animals directing towards ROIs.

    :param str config_path: Path to SimBA project config file in Configparser format
    :param Optional[Union[str, os.PathLike]] data_path: Path to folder or file holding the data used to calculate ROI aggregate statistics. If None, then defaults to the `project_folder/csv/outlier_corrected_movement_location` directory of the SimBA project. Default: None.

    .. note::
       `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

       `Example expected output file <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

    :example:
    >>> test = DirectingROIAnalyzer(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
    >>> test.run()
    >>> test.save()
    """

    def __init__(self, config_path: Union[str, os.PathLike],
                 data_path: Optional[Union[str, os.PathLike]] = None):

        check_file_exist_and_readable(file_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        self.data_paths = read_data_paths(path=data_path, default=self.outlier_corrected_paths,  default_name=self.outlier_corrected_dir, file_type=self.file_type)
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(expected_file_path=self.roi_coordinates_path)
        if not self.check_directionality_viable()[0]:
            raise InvalidInputError(msg="Cannot compute directionality towards ROIs. The ear and nose data is tracked in the project", source=self.__class__.__name__)
        self.read_roi_data()
        self.direct_bp_dict = self.check_directionality_cords()

    def __format_direction_data(
        self,
        direction_data: np.ndarray,
        nose_arr: np.ndarray,
        roi_center: np.ndarray,
        animal_name: str,
        shape_name: str,
    ) -> pd.DataFrame:

        x_min = np.minimum(direction_data[:, 1], nose_arr[:, 0])
        y_min = np.minimum(direction_data[:, 2], nose_arr[:, 1])
        delta_x = abs((direction_data[:, 1] - nose_arr[:, 0]) / 2)
        delta_y = abs((direction_data[:, 2] - nose_arr[:, 1]) / 2)
        x_middle, y_middle = np.add(x_min, delta_x), np.add(y_min, delta_y)
        direction_data = np.concatenate(
            (y_middle.reshape(-1, 1), direction_data), axis=1
        )
        direction_data = np.concatenate(
            (x_middle.reshape(-1, 1), direction_data), axis=1
        )
        direction_data = np.delete(direction_data, [2, 3, 4], 1)
        bp_data = pd.DataFrame(
            direction_data, columns=["Eye_x", "Eye_y", "Directing_BOOL"]
        )
        bp_data["ROI_x"] = roi_center[0]
        bp_data["ROI_y"] = roi_center[1]
        bp_data = bp_data[["Eye_x", "Eye_y", "ROI_x", "ROI_y", "Directing_BOOL"]]
        bp_data.insert(loc=0, column="ROI", value=shape_name)
        bp_data.insert(loc=0, column="Animal", value=animal_name)
        bp_data.insert(loc=0, column="Video", value=self.video_name)
        bp_data = bp_data.reset_index().rename(columns={"index": "Frame"})
        bp_data = bp_data[bp_data["Directing_BOOL"] == 1].reset_index(drop=True)
        return bp_data

    @staticmethod
    @jit(nopython=True, fastmath=True)
    def ccw(roi_lines: np.array, eye_lines: np.array, shape_type: str):
        def calc(A, B, C):
            return (C[1] - A[1]) * (B[0] - A[0]) > (B[1] - A[1]) * (C[0] - A[0])

        results = np.full((eye_lines.shape[0], 4), -1)
        for i in prange(eye_lines.shape[0]):
            eye, roi = eye_lines[i][0:2], eye_lines[i][2:4]
            min_distance = np.inf

            if shape_type == "Circle":
                reversed_roi_lines = roi_lines[::-1]
                for j in prange(roi_lines.shape[0]):
                    dist_1 = np.sqrt((eye[0] - roi_lines[j][0]) ** 2 + (eye[1] - roi_lines[j][1]) ** 2)
                    dist_2 = np.sqrt((eye[0] - roi_lines[j][2]) ** 2 + (eye[1] - roi_lines[j][3]) ** 2)
                    if (dist_1 < min_distance) or (dist_2 < min_distance):
                        min_distance = min(dist_1, dist_2)
                        results[i] = reversed_roi_lines[j]

            else:
                for j in prange(roi_lines.shape[0]):
                    line_a, line_b = roi_lines[j][0:2], roi_lines[j][2:4]
                    center_x, center_y = (line_a[0] + line_b[0] // 2, line_a[1] + line_b[1] // 2)
                    if calc(eye, line_a, line_b) != calc(roi, line_a, line_b) or calc(eye, roi, line_a) != calc(eye, roi, line_b):
                        distance = np.sqrt((eye[0] - center_x) ** 2 + (eye[1] - center_y) ** 2)
                        if distance < min_distance:
                            results[i] = roi_lines[j]
                            min_distance = distance

        return results

    def __find_roi_intersections(self, bp_data: pd.DataFrame, shape_info: dict):

        eye_lines = bp_data[["Eye_x", "Eye_y", "ROI_x", "ROI_y"]].values.astype(int)
        roi_lines = None
        if shape_info["Shape_type"].lower() == ROI_SETTINGS.RECTANGLE.value:
            top_left_x, top_left_y = (shape_info["topLeftX"], shape_info["topLeftY"])
            bottom_right_x, bottom_right_y = (shape_info["Bottom_right_X"], shape_info["Bottom_right_Y"])
            top_right_x, top_right_y = top_left_x + shape_info["width"], top_left_y
            bottom_left_x, bottom_left_y = (bottom_right_x - shape_info["width"], bottom_right_y)
            roi_lines = np.array([[top_left_x, top_left_y, bottom_left_x, bottom_left_y],
                                  [bottom_left_x, bottom_left_y, bottom_right_x, bottom_right_y],
                                  [bottom_right_x, bottom_right_y, top_right_x, top_right_y],
                                  [top_right_x, top_right_y, top_left_x, top_left_y]])

        elif shape_info["Shape_type"].lower() == ROI_SETTINGS.POLYGON.value:
            roi_lines = np.full((shape_info["vertices"].shape[0], 4), np.nan)
            roi_lines[-1] = np.hstack((shape_info["vertices"][0], shape_info["vertices"][-1]))
            for i in range(shape_info["vertices"].shape[0] - 1):
                roi_lines[i] = np.hstack((shape_info["vertices"][i], shape_info["vertices"][i + 1]))

        elif shape_info["Shape_type"].lower() == ROI_SETTINGS.CIRCLE.value:
            center = shape_info[["centerX", "centerY"]].values.astype(int)
            roi_lines = np.full((2, 4), np.nan)
            roi_lines[0] = np.array([center[0], center[1] - shape_info["radius"], center[0], center[1] + shape_info["radius"]])
            roi_lines[1] = np.array([center[0] - shape_info["radius"], center[1], center[0] + shape_info["radius"], center[1]])

        return self.ccw(roi_lines=roi_lines, eye_lines=eye_lines, shape_type=shape_info["Shape_type"])

    def run(self):
        self.results = []
        for file_cnt, file_path in enumerate(self.data_paths):
            _, self.video_name, _ = get_fn_ext(file_path)
            video_timer = SimbaTimer(start=True)
            print(f"Analyzing ROI directionality in video {self.video_name}...")
            data_df = read_df(file_path=file_path, file_type=self.file_type)
            video_roi_dict, shape_names = slice_roi_dict_for_video(data=self.roi_dict, video_name=self.video_name)
            for animal_name, bps in self.direct_bp_dict.items():
                ear_left_arr = data_df[[bps["Ear_left"]["X_bps"], bps["Ear_left"]["Y_bps"]]].values
                ear_right_arr = data_df[[bps["Ear_right"]["X_bps"], bps["Ear_right"]["Y_bps"]]].values
                nose_arr = data_df[[bps["Nose"]["X_bps"], bps["Nose"]["Y_bps"]]].values
                for roi_type, roi_type_data in video_roi_dict.items():
                    for _, row in roi_type_data.iterrows():
                        roi_center = np.array([row["Center_X"], row["Center_Y"]])
                        roi_name = row["Name"]
                        direction_data = FeatureExtractionMixin.jitted_line_crosses_to_static_targets(left_ear_array=ear_left_arr, right_ear_array=ear_right_arr, nose_array=nose_arr, target_array=roi_center)
                        bp_data = self.__format_direction_data(direction_data=direction_data, nose_arr=nose_arr, roi_center=roi_center, animal_name=animal_name, shape_name=roi_name)

                        eye_roi_intersections = pd.DataFrame(
                            self.__find_roi_intersections(
                                bp_data=bp_data, shape_info=row
                            ),
                            columns=[
                                "ROI_edge_1_x",
                                "ROI_edge_1_y",
                                "ROI_edge_2_x",
                                "ROI_edge_2_y",
                            ],
                        )
                        self.results.append(
                            pd.concat([bp_data, eye_roi_intersections], axis=1)
                        )
            video_timer.stop_timer()
            print(f"ROI directionality analyzed in video {self.video_name}... (elapsed time: {video_timer.elapsed_time_str}s)")
        if len(self.results) == 0:
            raise NoDataError(msg=f'No ROI DATA exists for data files {self.data_paths}', source=self.__class__.__name__)
        self.results_df = pd.concat(self.results, axis=0)

    def save(self, path: Optional[Union[str, os.PathLike]] = None):
        if not hasattr(self, "results_df"):
            raise InvalidInputError(msg="Run the ROI direction analyzer before saving")
        if path is None:
            path = os.path.join(
                self.logs_path, f"ROI_directionality_summary_{self.datetime}.csv"
            )
        self.results_df.to_csv(path)
        stdout_success(
            msg=f"Detailed ROI directionality data saved in {path}",
            source=self.__class__.__name__,
        )


#
# test = DirectingROIAnalyzer(config_path=r"C:\troubleshooting\spontenous_alternation\project_folder\project_config.ini")
# test.run()
# test.save()



__author__ = "Simon Nilsson"

import itertools
import json
import os
from collections import defaultdict
from typing import Any, Dict, List, Optional, Union

import h5py
import numpy as np
import pandas as pd

from simba.data_processors.interpolate import Interpolate
from simba.data_processors.smoothing import Smoothing
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.utils.checks import (check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_int,
                                check_str, check_valid_lst)
from simba.utils.enums import Methods
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_all_videos_in_project,
                                    get_video_meta_data, write_df)


class SLEAPImporterSLP(ConfigReader, PoseImporterMixin):
    """
    Class for importing SLEAP pose-estimation data into a SimBA project.

    .. note::
      Importing SLEAP .SLP files into SimBA come at long runtimes. For fater runtimes, use
      :meth:`simba.pose_importers.sleap_h5_importer.SLEAPImporterH5` or :meth:`simba.pose_importers.sleap_csv_importer.SLEAPImporterCSV`

    :param str config_path: path to SimBA project config file in Configparser format
    :param str data_folder: Path to folder containing SLEAP data in `csv` format.
    :param List[str] id_lst: Animal names. This will be ignored in one animal projects and default to ``Animal_1``.
    :param Optional[Dict[str, str]] interpolation_setting: Dict defining the type and method to use to perform interpolation {'type': 'animals', 'method': 'linear'}.
    :param Optional[Dict[str, Union[str, int]]] smoothing_settings: Dictionary defining the pose estimation smoothing method {'time_window': 500, 'method': 'gaussian'}.

    Example
    ----------
    >>> slp_importer = SLEAPImporterSLP(project_path="MyConfigPath", data_folder=r'MySLPDataFolder', actor_IDs=['Mouse_1', 'Mouse_2'], interpolation_settings="Body-parts: Nearest", smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
    >>> slp_importer.run()

    References
    ----------
    .. [1] Pereira et al., SLEAP: A deep learning system for multi-animal pose tracking, `Nature Methods`,
           2022.

    """

    def __init__(self,
                 project_path: Union[str, os.PathLike],
                 data_folder:  Union[str, os.PathLike],
                 id_lst: List[str],
                 interpolation_settings: Optional[Dict[str, str]] = None,
                 smoothing_settings: Optional[Dict[str, Any]] = None):

        ConfigReader.__init__(self, config_path=project_path, read_video_info=False)
        PoseImporterMixin.__init__(self)
        check_if_dir_exists(in_dir=data_folder)
        check_valid_lst(data=id_lst, source=f'{self.__class__.__name__} id_lst', valid_dtypes=(str,))
        if interpolation_settings is not None:
            check_if_keys_exist_in_dict(data=interpolation_settings, key=['method', 'type'], name=f'{self.__class__.__name__} interpolation_settings')
            check_str(name=f'{self.__class__.__name__} interpolation_settings type', value=interpolation_settings['type'], options=('body-parts', 'animals'))
            check_str(name=f'{self.__class__.__name__} interpolation_settings method', value=interpolation_settings['method'], options=('linear', 'quadratic', 'nearest'))
        if smoothing_settings is not None:
            check_if_keys_exist_in_dict(data=smoothing_settings, key=['method', 'time_window'], name=f'{self.__class__.__name__} smoothing_settings')
            check_str(name=f'{self.__class__.__name__} smoothing_settings method', value=smoothing_settings['method'], options=('savitzky-golay', 'gaussian'))
            check_int(name=f'{self.__class__.__name__} smoothing_settings time_window', value=smoothing_settings['time_window'], min_value=1)


        self.interpolation_settings, self.smoothing_settings = interpolation_settings, smoothing_settings
        self.data_folder, self.id_lst = data_folder, id_lst
        self.import_log_path = os.path.join(self.logs_path, f"data_import_log_{self.datetime}.csv")
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir)
        self.input_data_paths = self.find_data_files(dir=self.data_folder, extensions=[".slp"])
        self.data_and_videos_lk = self.link_video_paths_to_data_paths(data_paths=self.input_data_paths, video_paths=self.video_paths)
        if self.pose_setting is Methods.USER_DEFINED.value:
            self.__update_config_animal_cnt()
        if self.animal_cnt > 1:
            self.check_multi_animal_status()
            self.animal_bp_dict = self.create_body_part_dictionary(self.multi_animal_status, self.id_lst, self.animal_cnt, self.x_cols, self.y_cols, self.p_cols, self.clr_lst)
            if self.pose_setting is Methods.USER_DEFINED.value:
                self.update_bp_headers_file(update_bp_headers=True)
        print(f"Importing {len(list(self.data_and_videos_lk.keys()))} file(s)...")

    def __h5_to_dict(self, name, obj):
        attr = list(obj.attrs.items())
        if name == "metadata":
            jsonList = attr[1][1]
            jsonList = jsonList.decode("utf-8")
            final_dictionary = json.loads(jsonList)
            final_dictionary = dict(final_dictionary)
            return final_dictionary

    def __check_that_all_animals_exist_in_frame(self):
        existing_animals = list(self.frame_dict.keys())
        missing_animals = [x for x in range(self.animal_cnt) if x not in existing_animals]
        for missing_animal in missing_animals:
            self.frame_dict[missing_animal] = ([0] * ((len(self.analysis_dict["ordered_bps"]))) * 3)

    def __fill_missing_indexes(self):
        missing_indexes = list(set(list(range(0, self.video_info["frame_count"]))) - set(list(self.data_df.index)))
        missing_df = pd.DataFrame(0, index=missing_indexes, columns=self.analysis_dict["xyp_headers"])
        self.data_df = pd.concat([self.data_df, missing_df], axis=0)

    def run(self):
        self.analysis_dict, self.save_paths_lst = defaultdict(list), []
        for file_cnt, (video_name, video_data) in enumerate(self.data_and_videos_lk.items()):
            print(f"Analysing {video_name}...")
            video_timer = SimbaTimer(start=True)
            self.video_name = video_name
            in_h5 = h5py.File(video_data["DATA"], "r")
            self.sleap_dict = in_h5.visititems(self.__h5_to_dict)
            self.video_info = get_video_meta_data(video_path=video_data["VIDEO"])
            self.analysis_dict["bp_names"] = []
            self.analysis_dict["ordered_ids"] = []
            self.analysis_dict["ordered_bps"] = []
            self.analysis_dict["xy_headers"] = []
            self.analysis_dict["xyp_headers"] = []
            self.analysis_dict["animals_in_each_frame"] = []
            for bp in self.sleap_dict["nodes"]:
                self.analysis_dict["bp_names"].append(bp["name"])
            for orderVar in self.sleap_dict["skeletons"][0]["nodes"]:
                self.analysis_dict["ordered_ids"].append((orderVar["id"]))
            for index in self.analysis_dict["ordered_ids"]:
                self.analysis_dict["ordered_bps"].append(self.analysis_dict["bp_names"][index])

            with h5py.File(video_data["DATA"], "r") as file:
                self.analysis_dict["frames"] = file["frames"][:]
                self.analysis_dict["instances"] = file["instances"][:]
                self.analysis_dict["predicted_points"] = np.reshape(file["pred_points"][:], (file["pred_points"][:].size, 1))
            self.analysis_dict["no_frames"] = len(self.analysis_dict["frames"])

            for c in itertools.product(self.id_lst, self.analysis_dict["ordered_bps"]):
                x, y, p = (
                    str("{}_{}_x".format(c[0], c[1])),
                    str("{}_{}_y".format(c[0], c[1])),
                    (str("{}_{}_p".format(c[0], c[1]))),
                )
                self.analysis_dict["xy_headers"].extend((x, y))
                self.analysis_dict["xyp_headers"].extend((x, y, p))
            #
            self.data_df = pd.DataFrame(columns=self.analysis_dict["xyp_headers"])
            frames_lst = [l.tolist() for l in self.analysis_dict["frames"]]
            self.analysis_dict["animals_in_each_frame"] = [x[4] - x[3] for x in frames_lst]
            self.__create_tracks()
            if self.animal_cnt > 1:
                self.initialize_multi_animal_ui(animal_bp_dict=self.animal_bp_dict, video_info=self.video_info, data_df=self.data_df, video_path=video_data["VIDEO"])
                self.multianimal_identification()
            else:
                self.out_df = self.insert_multi_idx_columns(df=self.data_df.fillna(0))
            self.save_path = os.path.join(os.path.join(self.input_csv_dir, f"{self.video_name}.{self.file_type}"))



            write_df(df=self.out_df, file_type=self.file_type, save_path=self.save_path, multi_idx_header=True)
            if self.interpolation_settings is not None:
                interpolator = Interpolate(config_path=self.config_path, data_path=self.save_path, type=self.interpolation_settings['type'], method=self.interpolation_settings['method'], multi_index_df_headers=True, copy_originals=False)
                interpolator.run()
            if self.smoothing_settings is not None:
                smoother = Smoothing(config_path=self.config_path, data_path=self.save_path, time_window=self.smoothing_settings['time_window'], method=self.smoothing_settings['method'], multi_index_df_headers=True, copy_originals=False)
                smoother.run()
            video_timer.stop_timer()
            stdout_success(msg=f"Video {video_name} data imported...", elapsed_time=video_timer.elapsed_time_str)
        self.timer.stop_timer()
        stdout_success(msg="All SLEAP SLP data files imported", elapsed_time=self.timer.elapsed_time_str)

    def __create_tracks(self):
        start_frame = 0
        for frame_cnt, frame in enumerate(range(self.analysis_dict["no_frames"])):
            frame_idx = self.analysis_dict["frames"][frame_cnt][2]
            self.frame_dict = {}
            print(
                f"Restructuring SLEAP frame: {frame_cnt}/{self.analysis_dict['no_frames']}, Video: {self.video_name}"
            )
            self.cnt_animals_frm = self.analysis_dict["animals_in_each_frame"][frame]
            if self.cnt_animals_frm == 0:
                self.frame_dict[0] = [0] * len(self.analysis_dict["xyp_headers"])
                end_frame = start_frame + (
                    len(self.analysis_dict["ordered_bps"]) * self.cnt_animals_frm
                )

            else:
                end_frame = start_frame + (
                    len(self.analysis_dict["ordered_bps"]) * self.cnt_animals_frm
                )
                start_animal, end_animal = 0, len(self.analysis_dict["ordered_bps"])
                frame_arr = self.analysis_dict["predicted_points"][
                    start_frame:end_frame
                ]
                for instance_counter, animal in enumerate(range(self.cnt_animals_frm)):
                    currRow = []
                    animal_arr = frame_arr[start_animal:end_animal]
                    track_id = self.analysis_dict["instances"][instance_counter][4]
                    for bp in animal_arr:
                        currRow.extend((bp[0][0], bp[0][1], bp[0][4]))
                    self.frame_dict[track_id] = currRow
                    start_animal += len(self.analysis_dict["ordered_bps"])
                    end_animal += len(self.analysis_dict["ordered_bps"])

            if self.animal_cnt > 1:
                self.__check_that_all_animals_exist_in_frame()
            frame_lst = [
                item for sublist in list(self.frame_dict.values()) for item in sublist
            ]
            start_frame = end_frame
            try:
                self.data_df.loc[frame_idx] = frame_lst
            except ValueError:
                break

        self.data_df.fillna(0, inplace=True)
        self.__fill_missing_indexes()
        self.data_df.sort_index(inplace=True)
        self.data_df = self.insert_column_headers_for_outlier_correction(data_df=self.data_df, new_headers=self.bp_headers, filepath=self.video_name)


# test = SLEAPImporterSLP(project_path=r"C:\troubleshooting\sleap_two_animals_open_field\project_folder\project_config.ini",
#                         data_folder=r"C:\troubleshooting\sleap_two_animals_open_field\slp_import",
#                         id_lst=['Jarryd', 'Nilsson'],
#                         interpolation_settings=None,
#                         smoothing_settings = None) #Savitzky Golay
# test.run()








# test = SLEAPImporterSLP(project_path=r"C:\troubleshooting\slp_single_rat\project_folder\project_config.ini",
#                         data_folder=r"C:\troubleshooting\slp_single_rat\slp",
#                         id_lst=['Jarryd'],
#                         interpolation_settings=None,
#                         smoothing_settings = None) #Savitzky Golay
# test.run()


# test = SLEAPImporterSLP(project_path="/Users/simon/Desktop/envs/simba/troubleshooting/sleap_two_animals/project_folder/project_config.ini",
#                         data_folder=r'/Users/simon/Desktop/envs/simba/troubleshooting/sleap_two_animals/slp_import',
#                         id_lst=['Simon', 'JJ'],
#                         interpolation_settings={'type': 'animals', 'method': 'linear'},
#                         smoothing_settings = {'time_window': 500, 'method': 'gaussian'}) #Savitzky Golay
# test.run()
# #
# print('All SLEAP imports complete.')


# test = SLEAPImporterSLP(project_path="/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/data',
#                    id_lst=['Simon', 'Nastacia', 'JJ', 'Sam', 'Liana'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}}) #Savitzky Golay
# test.run()
# if test.animals_no > 1:
#     test.visualize_sleap()
# test.save_df()
# test.perform_interpolation()
# test.perform_smothing()
# print('All SLEAP imports complete.')


# test = ImportSLEAP(project_path="/Users/simon/Desktop/envs/troubleshooting/sleap_cody/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_cody/import_data',
#                    actor_IDs=['Animal_1', 'Animal_2'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}}) #Savitzky Golay
# test.initate_import_slp()
# if test.animals_no > 1:
#     test.visualize_sleap()
# test.save_df()
# test.perform_interpolation()
# test.perform_smothing()
# print('All SLEAP imports complete.')
